# -*- coding: utf-8 -*-
import unittest
import os
from io import StringIO, BytesIO

import numpy as np
import h5py

from test_data import *
from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.region_counts import get_extended_boundaries,\
                                           find_region_counts

from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *



###########################################

TRANSCRIPT_LENGTHS=\
"""GAPDH    90
VEGFA   12
FLT 40
MYC 1
P53 85"""


TRANSCRIPT_ANNOTATION=\
"""GAPDH    0   20  UTR5    0   -
GAPDH   20  50    CDS 0   -
GAPDH   50    90 UTR3   0   -
VEGFA   0   10 CDS 0   +
VEGFA   10 12 UTR3    0   +
FLT   0 15 UTR5    0   +
FLT   15 40 CDS    0   +
MYC 0  1    CDS 0   -
MYC 1    1    UTR3    0   -
P53 0   20  UTR5    0   +
P53 20  55  CDS 0   +
P53 55  85  UTR3    0   +"""


READ_SET_1=\
"""GAPDH 2  4  read_UTR5_1 0   +
GAPDH 3  5  read_UTR5_2 0   +
GAPDH 10  14  read_UTR5_3 0   +
GAPDH 10  14  read_UTR5_4 0   +
GAPDH 11  14  read_UTR5_5 0   +
GAPDH 12  14  read_UTR5_6 0   +
GAPDH 14  25  read_UTR5_7 0   +
GAPDH 15  20  read_UTR5_junc_1 0   +
GAPDH 16  20  read_UTR5_junc_2 0   +
GAPDH 17  20  read_UTR5_junc_3 0   +
GAPDH 20  25  read_UTR5_junc_4 0   +
GAPDH 23  25  read_UTR5_junc_5 0   +
GAPDH 24  30  read_CDS_1 0   +
GAPDH 30  35  read_CDS_2 0   +
GAPDH 44  46  read_CDS_3 0   +
GAPDH 45  55  read_UTR3_junc_1 0   +
GAPDH 53  57  read_UTR3_junc_2 0   +
GAPDH 54  57  read_UTR3_1 0   +
VEGFA 0  1  read_UTR5_junc_1 0  +
VEGFA 3  5  read_UTR5_junc_2 0  +
VEGFA 4  11  read_CDS_1 0  +
VEGFA 4  5  read_CDS_2 0  +
VEGFA 11  12  read_UTR3_junc_1 0  +
FLT 36  40  read_UTR3_junc_1 0 +
MYC 0  1  read_UTR5_junc_1 0  +
P53 5  15  read_UTR5_1 0  +
P53 25  35  read_CDS_1 0  +"""

#########################################
#####  Expected EXTENDED ANNOTATION 

GAPDH_regions = ( (0, 15), (15, 24), (24, 45), (45, 54), (54, 90) )
VEGFA_regions = ( (0, 0), (0, 4), (4, 5), (5, 12), (12, 12) )
FLT_regions   = ( (0, 10), (10, 19), (19, 35), (35, 40), (40, 40) )
MYC_regions   = ( (0, 0), (0, 1), (1, 1), (1, 1), (1, 1) )
P53_regions   = ( (0, 15), (15, 24), (24, 50), (50, 59), (59, 85) ) 
extended_boundary_regions = ( GAPDH_regions, VEGFA_regions,
                              FLT_regions, MYC_regions, P53_regions )
#########################################

#########################################

### Expected Region Counts
# Order is UTR5, UTR5_junc, CDS, UTR3_junc, UTR3

GAPDH_counts = ( 7, 5, 3, 2, 1 )
VEGFA_counts = ( 0, 2, 2, 1, 0 )
FLT_counts = ( 0, 0, 0, 1, 0 )
MYC_counts = ( 0, 1, 0, 0, 0 )
P53_counts = ( 1, 0, 1, 0, 0 )

expected_counts = ( GAPDH_counts, VEGFA_counts, FLT_counts,
                    MYC_counts, P53_counts )

#########################################


LEFT_SPAN  = 5
RIGHT_SPAN = 3

ANNOTATION = [\
               ([0, 20], [20, 50], [50, 90]),\
               ([0, 0], [0, 10], [10, 12]),\
               ([0, 15], [15, 40], [40, 40]),\
               ([0, 0], [0, 1], [1, 1]),\
               ([0, 20], [20, 55], [55, 85])
              ]


def _get_transcripts( file_in_string ):
    rows        = file_in_string.split("\n")
    pairs       = tuple( map( lambda x: x.split(), rows ) )
    ref_names   = tuple( map( lambda x: x[0], pairs ) )
    ref_lengths = tuple( map( lambda x: x[1], pairs ) )
    ref_lengths = tuple (map( int, ref_lengths ) )
    return (ref_names, ref_lengths)


class TestCreate(unittest.TestCase):

    def setUp(self):
        self.ref_len_file    = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file = StringIO(TRANSCRIPT_ANNOTATION)
        self.alignment_file  = StringIO(READ_SET_1)
        self.handle          = h5py.File(BytesIO())
        

    def tearDown(self):
        self.handle.close()


    def test_get_extended_boundaries(self):
        input_stream = StringIO( READ_SET_1 )
        ref_names , ref_lengths = _get_transcripts(TRANSCRIPT_LENGTHS)

        coverage = find_coverage(input_stream,  ref_names , ref_lengths )
        extended_boundaries = \
              get_extended_boundaries( annotation = ANNOTATION, 
                                       left_span  = LEFT_SPAN, 
                                       right_span = RIGHT_SPAN )

        for computed, expected in \
              zip(extended_boundaries , extended_boundary_regions):
            self.assertTrue( computed == expected ) 

  
    def test_find_region_counts(self):
        input_stream = StringIO(READ_SET_1)
        ref_names , ref_lengths = _get_transcripts(TRANSCRIPT_LENGTHS)

        coverage = find_coverage(input_stream,  ref_names , ref_lengths )
        rg = find_region_counts( coverage   = coverage, 
                                 annotation = ANNOTATION, 
                                 left_span  = LEFT_SPAN, 
                                 right_span = RIGHT_SPAN)

        for i in range(5):
            comparison = (rg[i,: ] == expected_counts[i] )
            self.assertTrue(np.all(comparison) ) 


if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-
from collections import OrderedDict

import numpy as np

#####################################################
# C O M M O N    N A M I N G   C O N V E N T I O N S
#####################################################
EXPERIMENTS_name = "experiments"
REFERENCE_name = "reference"

UTR5_name = "UTR5"
UTR3_name = "UTR3"
UTR5_JUNCTION_name = "UTR5_junction"
UTR3_JUNCTION_name = "UTR3_junction"
CDS_name = "CDS"

REGION_names          = [UTR5_name , CDS_name, UTR3_name ]
EXTENDED_REGION_names = [UTR5_name , UTR5_JUNCTION_name , 
                         CDS_name  , UTR3_JUNCTION_name , 
                         UTR3_name ]

METAGENE_name        = "metagene"
METAGENE_RADIUS_name = "metagene_radius"

LEFT_SPAN_name     = "left_span"
RIGHT_SPAN_name    = "right_span"

REF_DG_REFERENCE_NAMES   = "reference_names"
REF_DG_REFERENCE_LENGTHS = "reference_lengths"
REF_DG_START_SITE_COV    = "start_site_coverage"
REF_DG_STOP_SITE_COV     = "stop_site_coverage"
REF_DG_REGION_COUNTS     = "region_counts"
REF_DG_COVERAGE          = "coverage"

REF_ANNOTATION_NAME = "annotation"

LENGTH_MIN_name = "length_min"
LENGTH_MAX_name = "length_max"

RNASEQ_name = "rnaseq"


#####################################################
# H D F 5    S E T T I N G S
#####################################################
DEFAULT_COMPRESSION = "gzip"
DEFAULT_FLETCHER32  = True

SITE_COVERAGE_DT       = np.uint32
TRANSCRIPT_COVERAGE_DT = np.uint16
DEFAULT_COUNT_DT       = np.uint32
RNASEQ_DT              = np.float32

#####################################################
# B E D   C O L U M N   N A M E S
#####################################################
BED_CHROM  = "chrom"
BED_START  = "start"
BED_END    = "end"
BED_NAME   = "name"
BED_SCORE  = "score"
BED_STRAND = "strand"

BED_COLUMN_NAMES = [BED_CHROM,
                    BED_START, BED_END,
                    BED_NAME, BED_SCORE,
                    BED_STRAND ]

################################################################
# D A T A F R A M E   C O L U M N   N A M E S & S E T T I N G S
################################################################
DF_CSV_SEPARATOR = ","

DF_TRANSCRIPT   = "transcript"
DF_LIBRARY_NAME = "experiment_name"
DF_READLENGTH   = "read_length"

#####################################################
# M E T A D A T A
#####################################################
# Standard metadata for ribo files

ATTRS_REFERENCE       = "reference"
ATTRS_VERSION         = "ribogadgets_version"
ATTRS_FORMAT_VERSION  = "format_version"
ATTRS_TIME            = "time"
ATTRS_LENGTH_MIN      = "length_min"
ATTRS_LENGTH_MAX      = "length_max"
ATTRS_METAGENE_RADIUS = "metagene_radius"
ATTRS_LEFT_SPAN       = "left_span"
ATTRS_RIGHT_SPAN      = "right_span"
ATTRS_TOTAL_READS     = "total_reads"

ATTRS_ESSENTIAL_COMPATIBILITY = (ATTRS_REFERENCE, 
                                 ATTRS_FORMAT_VERSION,
                                 ATTRS_LENGTH_MIN, 
                                 ATTRS_LENGTH_MAX,
                                 ATTRS_METAGENE_RADIUS,
                                 ATTRS_LEFT_SPAN, 
                                 ATTRS_RIGHT_SPAN)

# This determines how the attrributes are displayed
# when information of rthe ribo file is requested
RIBO_METADATA_FOR_DISPLAY = OrderedDict(
    [ 
    (ATTRS_FORMAT_VERSION, "Ribo File Version"),
    (ATTRS_REFERENCE,      "Ribogadgets Version"),
    (ATTRS_TIME,           "Creation Time"),
    (ATTRS_REFERENCE,      "Reference"),
    (ATTRS_LENGTH_MIN,     "Min Read Length"),
    (ATTRS_LENGTH_MAX,     "Max Read Length"),
    (ATTRS_METAGENE_RADIUS,"Metagene Radius"),
    (ATTRS_LEFT_SPAN,      "Left Span"),
    (ATTRS_RIGHT_SPAN,     "Right Span") 
    ]
)

# The following metadata must match
# when merging experiments into one ribo file
RIBO_ESSENTIAL_METADATA =\
{
    ATTRS_METAGENE_RADIUS:\
        "number of nucleotides to the right "
        "and left of the start / stop sites",
    ATTRS_LEFT_SPAN: \
        "For 5/3 prime junction regions: "
        "number of nucleotides to the left of start / stop site",
    ATTRS_RIGHT_SPAN:\
        "For 5/3 prime junction regions: "
        "number of nucleotides to the right of start / stop site"
}

# USER METADATA is stored under this name
# in the attributes of the ribo file or the experiment
USER_METADATA = "metadata"

# Some essential metadata 
# The values correspond to display names
"""
LIBRARY_ESSENTIAL_METADATA = OrderedDict(\
    [ ("cell_line", "Cell Line"),
      ("sra", "SRA Number"),
      ("link", "External Link"),
      ("digestion_enzyme", "Digestion Enzyme"),
      ("digestion_duration", "Digestion Duration"),
      ("digestion_enzyme_volume", "D. Enz. Volume"),
      ("exp_prep_method", "Library Prep. Method"),
      ("chip", "Chip"),
      ("notes", "Notes:")]
)
"""

# All other metadata will be stored in "metadata" attribute of the
# sample in json format


#####################################################
# P L O T 
#####################################################

PLOT_COLORS = ["blue",  "red",    "green", 
               "brown", "orange", "violet", "black"]




# -*- coding: utf-8 -*-
import unittest
from io import StringIO, BytesIO
import os

import pandas as pd
import numpy as np
import h5py

from test_data import *
from ribogadgets import create
from ribogadgets.settings import *
from ribogadgets.rnaseq import get_rnaseq
from ribogadgets.core.exceptions import *


###########################################################

len_min = 5
len_max = 35

class TestCreate(unittest.TestCase):

    def setUp(self):
        self.ref_len_file     = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file  = StringIO(TRANSCRIPT_ANNOTATION)
        self.annotation_lines = StringIO(TRANSCRIPT_ANNOTATION).readlines()
        self.alignment_file   = StringIO(ALIGNMENT_FILE_1)
        self.handle           = h5py.File(BytesIO())
        
    def tearDown(self):
        self.handle.close()


    def test_init(self):
        create.initialize(self.handle, "merzifon", 
                          reference_name = "appris_human_v2")
        self.assertIn("merzifon", 
                      self.handle[EXPERIMENTS_name].keys())
        self.assertIn(REFERENCE_name, self.handle.keys())
        
    def test_set_reference_names_and_lengths(self):
        create.initialize(self.handle, "merzifon", 
                          reference_name = "appris_human_v2")
        create.set_reference_names_and_lengths(self.handle, self.ref_len_file)

        ref_h     = self.handle[REFERENCE_name]
        ref_names = ref_h[REF_DG_REFERENCE_NAMES][...].astype(str) 

        self.assertIn( "GAPDH",  ref_names[0])
        self.assertIn( "BRCA",  ref_names[3])
        self.assertEqual(4, len(ref_names) )

        ref_lengths = ref_h[REF_DG_REFERENCE_LENGTHS][...]
        self.assertEqual(875,     ref_lengths[1])
        self.assertEqual(1462,    ref_lengths[2])
        self.assertNotEqual(1462, ref_lengths[0])

    def test_create_ribo_file(self):
        test_ribo_file = "test_create_file.ribo"
        create.create_ribo_file(test_ribo_file, "iconia", 
                                alignment_file  = self.alignment_file,
                                reference_name  = "appris_human_v2",
                                length_min      = len_min,
                                length_max      = len_max,
                                metagene_radius = 3,
                                left_span       = 3, 
                                right_span      = 2,
                                lengths_file    = self.ref_len_file,
                                annotation_file = self.annotation_file)
        try:
            with h5py.File(test_ribo_file) as h:
                self.assertEqual( tuple(h[EXPERIMENTS_name].keys())[0] , "iconia")
        finally:
            os.remove(test_ribo_file)


    def test_get_reference_names(self):
        create.initialize(self.handle, "merzifon", 
                          reference_name = "appris_human_v2")
        create.set_reference_names_and_lengths(self.handle, self.ref_len_file)

        create.set_annotation( h5_handle        = self.handle, 
                               annotation_lines = self.annotation_lines )

        ref_array = self.handle[REFERENCE_name][REF_ANNOTATION_NAME][...]
        self.assertEqual( ref_array[0,0], 50 )
        self.assertEqual( ref_array[1,0], 0 )
        self.assertEqual( ref_array[1,2], 875 )
        self.assertEqual( ref_array[2,1], 1251 )
        self.assertEqual( ref_array[3,2], 565 )


    def test_set_coverage_vectors(self):

        create.initialize(self.handle, "merzifon", 
                          reference_name = "appris_human_v2")

        create.set_reference_names_and_lengths(self.handle, self.ref_len_file)
        create.set_annotation( h5_handle        = self.handle, 
                               annotation_lines = self.annotation_lines )

        create.set_coverage_vectors(self.handle, 5)

        start_site_cov = self.handle[REFERENCE_name][REF_DG_START_SITE_COV][...]
        stop_site_cov  = self.handle[REFERENCE_name][REF_DG_STOP_SITE_COV][...]
        self.assertListEqual( list(start_site_cov) , [2,2,2,3,3,4,4,4,4,4,4] )
        self.assertListEqual( list(stop_site_cov), [4,4,4,4,4,4,4,4,4,4,3] )
        self.assertTrue( list(stop_site_cov) != [6,4,4,4,4,4,4,4,4,4,3] )

    def test_rnaseq(self):
        test_ribo_file      = "test_create_file.ribo"
        self.rnaseq_handle = StringIO(RNASEQ_DATA_1)


        create.create_ribo_file(test_ribo_file, "iconia", 
                                alignment_file  = self.alignment_file,
                                reference_name  = "appris_human_v2",
                                length_min      = len_min,
                                length_max      = len_max,
                                metagene_radius = 3,
                                left_span       = 3, 
                                right_span      = 2,
                                rnaseq_file     = self.rnaseq_handle,
                                lengths_file    = self.ref_len_file,
                                annotation_file = self.annotation_file)
        try:
            with h5py.File(test_ribo_file) as h:
                rnaseq_df = get_rnaseq(h, "iconia")
                self.assertEqual(rnaseq_df.index[1], "VEGFA" )
                self.assertTrue(  np.isclose(rnaseq_df.loc["VEGFA", "iconia"], 15.46)  )
        finally:
            os.remove(test_ribo_file)

    def test_invalid_name_check(self):
        test_ribo_file = "test_create_file.ribo"

        with self.assertRaises(RiboBaseError) as exception_context:

            create.create_ribo_file(test_ribo_file, 
                                    experiment_name = "this,that", 
                                    alignment_file  = self.alignment_file,
                                    reference_name  = "appris_human_v2",
                                    length_min      = len_min,
                                    length_max      = len_max,
                                    metagene_radius = 3,
                                    left_span       = 3, 
                                    right_span      = 2,
                                    lengths_file    = self.ref_len_file,
                                    annotation_file = self.annotation_file)

        if os.path.isfile(test_ribo_file):
            os.remove(test_ribo_file)

if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-

import gzip
import io
from io import StringIO, IOBase
import sys
import pipes
import shutil

from ribogadgets.core.exceptions import RiboBaseError

#############################################

def get_file_opener_by_extension(file_path):
    file_string = file_path.lower()
    if file_string.endswith("gz") or \
       file_string.endswith("gzip"):

       return gzip.open
    else:
        return open

#####################################################################        

def open_by_extension(file_path, mode = "rt"):
    file_opener = get_file_opener_by_extension(file_path)
    return file_opener(file_path, mode)

#####################################################################

def read_all_lines(file_input):
    if issubclass( type(file_input), IOBase ):
        return file_input.readlines()

    elif type(file_input) == str:
        with open_by_extension( file_input ) as file_handle:
            all_lines = file_handle.readlines()
        return all_lines
    
    else:
        raise RiboBaseError("Input file must be a stream or a file path")

#####################################################################

def get_alignment_file_handle(alignment_file, alignment_format = "bed"):

    # Probably we dont need this
    # but an extra check doesnt hurt
    if alignment_format not in ("bed", "bam"):
        raise RiboBaseError("File format can be bam or bed.")

    # If the alignment file is bam
    # it needs to be converted to bed using bamToBed
    if alignment_format == "bam":
        if not shutil.which('bamToBed'):
            raise RiboBaseError("Ribogadgets could not "
                                "find the executable bamToBed."
                                "This executable is needed for bam files.")
        pipe_t = pipes.Template()

    if not alignment_file:
        # ALignment data is coming from standard input
        if alignment_format == "bed":
            alignment_file_handle = sys.stdin
        else:
            # We need to convert bam to bed using
            # bamToBed through pipes
            pipe_t.append("bamToBed " , '--')
            alignment_file_handle = pipe_t.open(None, 'r')
            
    else:
        # There is an actual alignment file
        # that needs to be opened properly
        if alignment_format == 'bam':
            pipe_t.prepend("bamToBed -i {}".format(alignment_file), ".-")
            alignment_file_handle = pipe_t.open(None, 'r')
        else:
            alignment_file_handle = flex_open(alignment_file, 'r')


    return alignment_file_handle


#####################################################################

def flex_open(file, mode = "rt"):
    file_type = type(file)

    if issubclass(file_type, IOBase):
        return file
    elif file_type == str:
        return open_by_extension(file, mode)

#####################################################################

def flex_out(file, mode = "wt"):
    if not file:
        return sys.stdout
    else:
        return flex_open(file , mode)








# -*- coding: utf-8 -*-

from functools import partial, wraps
from collections import OrderedDict

import h5py
import pandas as pd
import numpy as np

from .settings import *
from .core.get_gadgets import *
from .core.aggregate_by_length import aggregate_region_counts,\
                                      aggregate_start_site,\
                                      aggregate_stop_site

from .core.verify import verify_extended_region_name,\
                         verify_site_type

from .core.exceptions import *

from .io.file import flex_out
from .io.dump_coverage import print_coverage

from .core.verify import make_cli_function, \
                         check_experiment_list_in_ribo_handle, \
                         check_length_range

######################################################################                         

COLUMN_INDECES = dict( zip(EXTENDED_REGION_names, 
                           tuple(range(0, len(EXTENDED_REGION_names)))))

#######################################################################

#######################################################################

@check_experiment_list_in_ribo_handle
@check_length_range
def dump_metagene( ribo_handle, *,
                   site_type, 
                   sum_lengths, 
                   sum_references,
                   range_lower     = 0, 
                   range_upper     = 0,
                   experiment_list = [] ):
    """
    **TODO:**
    Complete documentation
    """

    length_min, length_max = get_read_length_range(ribo_handle)

    if not range_lower:
        range_lower = length_min
    if not range_upper:
        range_upper = length_max

    verify_site_type(site_type)

    if not experiment_list:
        experiment_list = get_experiment_names(ribo_handle)

    if site_type == REF_DG_START_SITE_COV:
        get_site = aggregate_start_site
    elif site_type == REF_DG_STOP_SITE_COV:
        get_site = aggregate_stop_site
    else:
        raise ValueError("Wring site type "+ site_type +\
                         "It must be one of {} or {}".format(\
                           REF_DG_START_SITE_COV, REF_DG_STOP_SITE_COV ))

    metagene_data = get_site(ribo_handle     = ribo_handle, 
                             range_lower     = range_lower, 
                             range_upper     = range_upper,
                             sum_values      = sum_lengths,
                             experiment_list = experiment_list)

    metagene_radius = get_metagene_radius(ribo_handle)

    # Note that the second value is excluded
    # so we add +1 tp include the +metagene_radius position
    coverage_interval = list(range( -1 * metagene_radius, 
                                     metagene_radius + 1 ) )
    reference_names = list(get_reference_names(ribo_handle))

    if sum_lengths:
        column_labels = coverage_interval
        create_df_pre = partial( pd.DataFrame, columns = column_labels)

        if sum_references:
            create_df = (lambda x: pd.DataFrame(
                               create_df_pre(x).sum()   ) )
        else:
            create_df = partial( pd.DataFrame, columns = column_labels,
                                 index = reference_names)
    else:
        column_labels = [DF_READLENGTH]  + coverage_interval
        create_df_pre = partial( pd.DataFrame, columns = column_labels)

        if sum_references:
            create_df = (lambda x: create_df_pre(x).\
                            groupby([DF_READLENGTH]).sum()  )
        else:
            reference_name_repetition_time = (range_upper - range_lower) + 1 
            reference_labels = \
                reference_names * reference_name_repetition_time
            create_df = partial( pd.DataFrame, columns = column_labels,
                                 index = reference_labels)

  
    metagene_dataframes = list(map(create_df , metagene_data))

    if sum_lengths and sum_references:
        metagene_dataframes = list(map(lambda x: x.T, 
                                 metagene_dataframes) )

    dataframes_dict = OrderedDict( 
                          tuple(zip(experiment_list, 
                                     metagene_dataframes)) )

    return dataframes_dict




def merge_metagene_dataframes(df_dict, 
                              sum_references, 
                              sum_lengths):

    """
    Incomplete documentation
    """

    experiment_list = tuple(df_dict.keys())

    if sum_references:
        names = [DF_LIBRARY_NAME]
    else:
        names = [DF_LIBRARY_NAME, DF_TRANSCRIPT]

    # We try to aoid the redundant index colum
    # if references and lengths are summed
    if sum_references and sum_lengths:
        return pd.concat( tuple(df_dict.values()), 
               keys  = experiment_list,
               names = names,
               join  = "outer").reset_index(level=1, drop=True)

    else:
        return pd.concat( tuple(df_dict.values()), 
               keys  = experiment_list,
               names = names,
               join  = "inner")

@make_cli_function
def dump_metagene_wrapper( ribo_file, 
                           output_file, 
                           site_type, 
                           sum_lengths, 
                           sum_references,
                           range_lower     = 0, 
                           range_upper     = 0,
                           experiment_list = [] ):
    """
    TODO:
    Complete documentation
    """

    if site_type.lower() == "start":
        site_type = REF_DG_START_SITE_COV
    elif site_type.lower() == "stop":
        site_type = REF_DG_STOP_SITE_COV
    else:
        raise ValueError("Wring site type:", site_type)


    ribo_handle = h5py.File(ribo_file, "r")

    try:
        df_dict = dump_metagene( ribo_handle     = ribo_handle,
                                 site_type       = site_type, 
                                 sum_lengths     = sum_lengths, 
                                 sum_references  = sum_references,
                                 range_lower     = range_lower, 
                                 range_upper     = range_upper,
                                 experiment_list = experiment_list ) 

        site_coverage = merge_metagene_dataframes(df_dict, 
                                 sum_references, sum_lengths)
    except ValueError as e:
        print(e)
        exit(1)
    finally:
        ribo_handle.close()

    if output_file:
        site_coverage.to_csv(output_file, sep = DF_CSV_SEPARATOR)
    else:
        print(site_coverage.to_csv( sep = DF_CSV_SEPARATOR) )


#########################################################################

@check_experiment_list_in_ribo_handle
@check_length_range
def dump_region(ribo_handle, *,
                region_name,
                sum_lengths, 
                sum_references, 
                range_lower      = 0 , 
                range_upper      = 0 ,
                experiment_list  = []):

    """
    Dumps a selection region counts to a dataframe.
    Available regions are:
       UTR5, UTR5_junction, CDS, UTR3_junction, UTR3

    For each experiment, there will be a column with the experiment name.

    Originally, we keep counts for all transcripts and all lengths
    The counts can be summed accross transcripts or accros lengths
    by setting the flags sum_lengths or sum)references.

    If the counts are NOT summed across lengths, then
    the firsy column will be for the corresponding read lengths

    If the sums are not across transcripts (references), then
    the index column will indicate the corresponding transcript (reference)
    name.

    Parameters
    ----------

    ribo_handle : h5py.File
        A pointer to open hdf5 file

    region_name : str
            Type of the region. 
            Choices are: UTR5, UTR5_junction, CDS, UTR3_junction, UTR3

    sum_lengths : Boolean
            sum the counts across the lengths

    sum_references : Boolean
            sum the counts across transcripts (references)

    range_lower : int
            starting read length to be taken

    range_upper : int
            ending read length to be taken

    experiment_list : list(str)
                   the experiments to be reported
                   If empty, all experiments will be taken.

    Returns
    -------

    result_df : A pandas dataframe

    """

    length_min, length_max = get_read_length_range(ribo_handle)

    if not range_lower:
        range_lower = length_min
    if not range_upper:
        range_upper = length_max

    verify_extended_region_name(region_name)

    if experiment_list:
        actual_experiment_list = experiment_list
    else:
        actual_experiment_list = get_experiment_names(ribo_handle)
    
    region_counts = aggregate_region_counts( 
                             ribo_handle     = ribo_handle, 
                             range_lower     = range_lower, 
                             range_upper     = range_upper, 
                             sum_values      = sum_lengths,
                             experiment_list = actual_experiment_list )



    
    # Note that the first column is for the read length
    # so we need to add 1 to the column index
    offset           =  0 if sum_lengths else 1 
    collected_counts = [ c[:, COLUMN_INDECES[region_name] + offset]
                            for c in region_counts ]

    reference_names  = get_reference_names(ribo_handle)
    df_index         = []
    column_labels    = []

    #### If we are NOT summing the counts by read lengths
    #### Then the counts are reported per read length
    #### So we need to keep them in the first column
    if not sum_lengths:
        read_lengths  = region_counts[0][:,0]
        df_contents   = [read_lengths] + collected_counts
        column_labels = [DF_READLENGTH] + list(actual_experiment_list)
    else:
        column_labels = actual_experiment_list
        df_contents   = collected_counts


    data_table = np.array( df_contents )
    counts_df  = pd.DataFrame(np.transpose(data_table),
                             columns = column_labels)

    if sum_references:
        if sum_lengths:
            result_df = pd.DataFrame(counts_df.sum())
        else:
            result_df = counts_df.groupby([DF_READLENGTH]).sum()
    else:
        if sum_lengths:
            reference_name_repetition_time = 1
            column_labels                  = actual_experiment_list
        else:
            reference_name_repetition_time = (range_upper - range_lower) + 1
            column_labels                  = \
                 [DF_READLENGTH] + list(actual_experiment_list)

        reference_labels = list(reference_names) * reference_name_repetition_time
        result_df        = pd.DataFrame(np.transpose(data_table), 
                                          index   = reference_labels,
                                          columns = column_labels)

    return result_df




@make_cli_function   
def dump_region_wrapper(ribo_file, 
                        output_file, 
                        region_name,
                        sum_lengths, 
                        sum_references, 
                        range_lower     = 0, 
                        range_upper     = 0,
                        experiment_list = []):

    """
    A wrapper function for the dump_region function.

    If output file is given, the resulting data frame will be saved
    in it. Else, the dataframe will be printed on the standard output
    in csv format.

    See dump_region function for details.
    """

    ribo_handle = h5py.File(ribo_file, "r")

    try:
        region_counts = dump_region( ribo_handle     = ribo_handle, 
                                     region_name     = region_name,
                                     sum_lengths     = sum_lengths, 
                                     sum_references  = sum_references, 
                                     range_lower     = range_lower , 
                                     range_upper     = range_upper ,
                                     experiment_list = experiment_list)
    except ValueError as e:
        print(e)
        exit(1)
    finally:
        ribo_handle.close()

    if output_file:
        region_counts.to_csv(output_file, sep = DF_CSV_SEPARATOR)
    else:
        print(region_counts.to_csv( sep = DF_CSV_SEPARATOR) )



##########################################################

def dump_annotation(ribo_handle):
    """
    Returns annotation of a ribo file in bed format
    in string form.

    Parameters
    ----------

    ribo_handle : h5py.File
         hdf5 handle for the ribo file


    Returns
    -------

    A string that can be output directly as a bed file.
    """

    boundaries = get_region_boundaries(ribo_handle)
    names      = get_reference_names(ribo_handle)

    bed_rows   = list()

    for ref_name, ref_boundaries in zip(names, boundaries):
        for region_name, region_boundaries in zip(REGION_names, ref_boundaries ):
            if region_boundaries[1] <= region_boundaries[0]:
                continue
            bed_entries = tuple( map( str, 
                                 [ref_name, 
                                  region_boundaries[0], region_boundaries[1],
                                  region_name, 0, "+"] ) )
            bed_rows.append( "\t".join(bed_entries) )
    return "\n".join(bed_rows)



@make_cli_function
def dump_annotation_wrapper(ribo_file, output_file = ""):
    """
    Dumps region annotation of a given ribo file.
    If the output file is not provided, 
    the result is printed on the standard output

    parmaters
    ---------
    ribo_file: ribo file path

    output_file: output file path
    """

    ribo_handle = h5py.File(ribo_file, "r")
    annotation = dump_annotation(ribo_handle)

    if output_file:
        with open(output_file, "w") as output_stream:
            print(annotation, file=output_stream)
    else:
            print(annotation)

    ribo_handle.close()

#####################################################################

def dump_reference_lengths(ribo_handle):
    """
    Dumps refrence names and their lengths in a given
    ribo file.

    The output is a pandas dataframe whose index is reference names

    parameters
    ----------
    ribo_file: h5py.File
       ribo handle

    Returns
    -------
    ref_lens_df: pandas.DataFrame
    """

    ref_lengths  = get_reference_lengths(ribo_handle)
    ref_names    = get_reference_names(ribo_handle)
    ref_lens_df  = pd.DataFrame(ref_lengths , 
                                 index   = ref_names,
                                 columns = ["length"], 
                                 dtype   = TRANSCRIPT_COVERAGE_DT) 
    return ref_lens_df


@make_cli_function
def dump_reference_lengths_wrapper(ribo_file, 
                                   output_file = "", sep = "\t"):
    """
    Dumps refrence names and their lengths in a given
    ribo file.

    If no output file is given, the result is printed to
    the standard output.

    parameters
    ----------
    ribo_file: ribo file path

    output_file: output file path
    """

    ribo_handle       = h5py.File(ribo_file, "r")
    ref_lengths_df    = dump_reference_lengths(ribo_handle)
    ref_lengths_str   = ref_lengths_df.to_csv(sep = sep, header = None)

    with flex_out(output_file) as output_stream:
        print(ref_lengths_str, file = output_stream)

    ribo_handle.close()


#####################################################################

@check_experiment_list_in_ribo_handle
@check_length_range
def dump_coverage( ribo_handle, *, 
                   experiment_name,
                   range_lower = 0, 
                   range_upper = 0):

    """
    Returns an ordered dictionary 
    whose keys are transcript names 
    and values are coverage of the transcript.

    The coverage values in the range
    range_lower and range_upper
    are summed up.
    So if N = range_lower = range_upper,
    then the coverage values at length N
    are reported.

    Arguments
    ---------
    ribo_handle: h5py.File
       Handle to open ribo file

    experiment_name: str
       Name of the experiment 
       whose coverage is to be extracted

    range_lower: int
       Minimum read length to be taken

    range_upper: int
       Maximum read length to be taken 

    Returns
    -------
    transcript_coverage: OrderedDict( np.array )
       The keys of transcript coverage are transcript names
       The values are numpy arrays giving the coverage.
       So the value at position i corresponds to coverage on
       the ith nucleotide.
    """

    if not has_coverage_data(ribo_handle, experiment_name):
        raise LibraryDoesntExist(
               "The experiment {} doesn't have coverage data"\
                .format(experiment_name))


    ref_lengths = get_reference_lengths(ribo_handle)
    ref_names   = get_reference_names(ribo_handle)

    min_length, max_length = \
        get_read_length_range(ribo_handle)

    total_nucleotides  = sum(ref_lengths)
    total_slice_array  = np.zeros( total_nucleotides, 
                                   dtype = TRANSCRIPT_COVERAGE_DT )
    length_offset      = range_lower - min_length
    coverage_handle    = ribo_handle[EXPERIMENTS_name]\
                                    [experiment_name][REF_DG_COVERAGE]

    for this_length in range(range_lower, range_upper + 1):
        start_index = (this_length - min_length) * total_nucleotides
        end_index   = start_index + total_nucleotides  

        total_slice_array += \
            coverage_handle[REF_DG_COVERAGE][start_index : end_index]

    transcript_coverage = OrderedDict()
    start_index         = 0
    end_index           = 0

    for t_name, t_length in zip(ref_names, ref_lengths):
        end_index                   = start_index + t_length 
        transcript_coverage[t_name] = \
              total_slice_array[start_index : end_index ]
        start_index += t_length

    return transcript_coverage

########################################################################


@make_cli_function
def dump_coverage_wrapper(ribo_file, 
                          output_file, 
                          experiment_name,
                          range_lower = 0, 
                          range_upper = 0,
                          file_format = "bg",
                          separator   = ","):

    """
    This is a wrapper function for dump_coverage.

    This function arranges the output files and
    calling appropriate io functions for writing
    the output.

    See the documentation of dump_coverage for details.
    """

    ribo_handle = h5py.File(ribo_file, "r")

    if file_format not in ["bg", "tsv"]:
        print("File format can be either \"bg\" or \"tsv\".")
        exit(1)

    if not has_coverage_data(ribo_handle, experiment_name):
        exit("Error: The experiment {} does not have coverage data."\
              .format(experiment_name))

    with flex_out(output_file, "wt") as outstream:

        coverage =  dump_coverage( 
                       ribo_handle     = ribo_handle, 
                       experiment_name = experiment_name,
                       range_lower     = range_lower, 
                       range_upper     = range_upper)

        print_coverage(coverage, 
                        file_handle = outstream,
                        file_format = file_format,
                        sep         = separator)

    ribo_handle.close()



# -*- coding: utf-8 -*-
import unittest
import os
from io import StringIO, BytesIO

import pandas as pd
import h5py

from test_data import *
from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths
from ribogadgets.settings import *


###########################################

TRANSCRIPT_LENGTHS=\
"""GAPDH    10
VEGFA   5
MYC 2"""


TRANSCRIPT_ANNOTATION=\
"""GAPDH    0   4  UTR5    0   -
GAPDH   4  7    CDS 0   -
GAPDH   7    10 UTR3   0   -
VEGFA   0   3 CDS 0   +
VEGFA   3 5 UTR3    0   +
MYC 0  1    CDS 0   -
MYC 1    2    UTR3    0   -"""

READ_SET_1=\
"""GAPDH 2  4  read_1 0   +
MYC 0  1  read_2 0 +
VEGFA 4  5  read_3 0 +
MYC 0  1  read_4 0 +
MYC 0  1  read_5 0 +
VEGFA 4  5  read_6 0 +
VEGFA 1  2  read_7 0 +"""

###########################################


def check_READ_SET_1(self, result):
    self.assertEqual( result["MYC"][0], 3 )
    self.assertEqual( result["VEGFA"][4], 2 )
    self.assertEqual( result["VEGFA"][1], 1 )
    self.assertNotEqual( result["VEGFA"][0], 2 )
    self.assertEqual( result["GAPDH"][2], 1 )    

def _get_transcripts(file_in_string):
    rows = file_in_string.split("\n")
    pairs = tuple( map( lambda x: x.split(), rows ) )
    ref_names = tuple( map( lambda x: x[0], pairs ) )
    ref_lengths = tuple( map( lambda x: x[1], pairs ) )
    ref_lengths = tuple (map( int, ref_lengths ) )
    return (ref_names, ref_lengths)


class TestCreate(unittest.TestCase):

    def setUp(self):
        self.ref_len_file     = StringIO( TRANSCRIPT_LENGTHS )
        self.annotation_file  = StringIO( TRANSCRIPT_ANNOTATION )
        self.alignment_file   = StringIO(READ_SET_1)

        self.handle = h5py.File(BytesIO())

    def tearDown(self):
        self.handle.close()

    def test_coverage_solo(self):
        input_stream = StringIO(READ_SET_1)
        ref_names , ref_lengths = _get_transcripts(TRANSCRIPT_LENGTHS)
        result = find_coverage( input_stream, ref_names , ref_lengths )
        check_READ_SET_1(self, result)

    def test_coverage_from_ribo(self):
        create.initialize(self.handle, "merzifon", 
                          reference_name = "appris_human_v2")
        create.set_reference_names_and_lengths(self.handle, self.ref_len_file)
        create.set_annotation( h5_handle        = self.handle, 
                               annotation_lines = TRANSCRIPT_ANNOTATION.split("\n") )

        input_stream = StringIO(READ_SET_1)
        ref_names    = get_reference_names(self.handle)
        ref_lengths  = get_reference_lengths(self.handle)
        result       = find_coverage( input_stream, ref_names , ref_lengths )
        check_READ_SET_1(self, result)
        

if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
from collections import OrderedDict
import os
from io import IOBase

import h5py
import yaml
import numpy as np
import pandas as pd

from .settings import *
from .io.file import flex_open, flex_out
from .core.get_gadgets import *
from .core.exceptions import *
from .core.verify import prompt_user, make_cli_function,\
                         check_experiment_list_in_ribo_handle
from ._version import (__format_version__,
                      __version__) 

###########################################################


class NORNASEQ(Exception):
    pass

###########################################################

@check_experiment_list_in_ribo_handle
def set_rnaseq(ribo_handle, name, rnaseq_df):
    """
    Set the rna-seq data

    Parameters
    ----------
    ribo_handle: h5py.File
        Open ribo file handle

    name: str
        Name of the experiment

    rnaseq_df: pd.DataFrame
        Index is the transcript names
        The first (and only) column has transcript expression values.

    Returns
    -------
    actual_rnaseq_df[0]: np.array
       Transcript expression array
    """

    rnaseq_datagroup  = ribo_handle[EXPERIMENTS_name][name]\
                           .require_group(RNASEQ_name)

    t_names = get_reference_names(ribo_handle)                       

    actual_rnaseq_df  = pd.DataFrame(np.zeros( len(t_names), 
                                               dtype = RNASEQ_DT ), 
                                      index = t_names )

    for transcript in rnaseq_df.index:
        actual_rnaseq_df.loc[transcript, 0] = rnaseq_df.loc[transcript, 0]

    rnaseq_dataset = rnaseq_datagroup.get(RNASEQ_name, None)

    if rnaseq_dataset:
        rnaseq_dataset[...] = actual_rnaseq_df.loc[: , 0]
    else:
        rnaseq_datagroup.create_dataset(
                                RNASEQ_name, 
                                shape       = (len(t_names),),
                                data        = actual_rnaseq_df[0],
                                dtype       = RNASEQ_DT,
                                compression = DEFAULT_COMPRESSION, 
                                fletcher32  = DEFAULT_FLETCHER32)
                                     
    return actual_rnaseq_df[0]

@make_cli_function
def set_rnaseq_wrapper(ribo_file, name, rnaseq_file, 
                       sep = "\t", force = True):
    """
    Wrapper for set_metadata.

    It makes sure that the provided transcript names
    are actually in the reference transcripts.

    See set_rnaseq for details,
    """
 
    with flex_open(rnaseq_file, "rt") as input_stream,\
         h5py.File(ribo_file, "r+") as ribo_handle:

        experiment_handle = ribo_handle[EXPERIMENTS_name][name]
        prompt_message    = "This will overwrite existing RNA-Seq data." 
        attr_exists       = \
           experiment_handle.get(RNASEQ_name , None) \
              and \
           experiment_handle[RNASEQ_name].get(RNASEQ_name, name)

        prompt_user( message     = prompt_message, 
                     attr_exists = attr_exists, 
                     force       = force)

        rnaseq_df = pd.read_csv(input_stream, 
                                    header    = None, 
                                    names     = [0], 
                                    index_col = 0, 
                                    sep       = sep )

        transcript_names = get_reference_names(ribo_handle)

        for t in rnaseq_df.index:
            if t not in transcript_names:
                print("Error: Invalid transcrpt name: {}".format(t))
                exit(1)

        set_rnaseq(ribo_handle = ribo_handle, 
                   name        = name, 
                   rnaseq_df   = rnaseq_df)
    



def _get_single_rna_seq(ribo_handle, name):
    """
    Helper function for get_rnaseq

    It return rnaseq data of a single experiment if exists
    """

    if not rnaseq_exists(ribo_handle, name):
        raise NORNASEQ("{} doesn't have RNA_seq data".format(name))

    rnaseq_np = ribo_handle[EXPERIMENTS_name]\
                    [name][RNASEQ_name][RNASEQ_name][...]
    ref_names = get_reference_names(ribo_handle)
    rnaseq_df = pd.DataFrame( rnaseq_np , 
                              index   = ref_names , 
                              columns = [name])

    return rnaseq_df


def _get_all_rna_seq(ribo_handle):
    """
    Helper function for get_rnaseq

    It return rnaseq data of a all experiments in the ribo file
    """

    has_rnaseq = False

    experiment_list = get_experiment_names(ribo_handle)
    rnaseq_df_list  = list()

    for experiment in experiment_list:
        if rnaseq_exists(ribo_handle, experiment):
            has_rnaseq = True
            rnaseq_df_list.append( _get_single_rna_seq(ribo_handle, experiment) )

    if not has_rnaseq:
        return None
    else:
        all_df = pd.concat(rnaseq_df_list, axis = 1)
        return all_df

@check_experiment_list_in_ribo_handle
def get_rnaseq(ribo_handle, name = None):
    """
    Returns rnaseq profile of the data if it exists

    Parameters
    ----------
    ribo_handle: h5py.File
        Open ribo file handle

    name: str
        Name of the experiment

    Returns
    -------
    rnaseq_df: pd.DataFrame 
        transcript expression table.
        Indices are transcript names
    """

    if not name:
        return _get_all_rna_seq(ribo_handle)
    else:
        return _get_single_rna_seq(ribo_handle, name)



@make_cli_function
def get_rnaseq_wrapper(ribo_file, name, output, sep = "\t"):
    """
    Wrapper for get_rnaseq
    """

    ribo_handle = h5py.File(ribo_file , "r")

    try:
        rnaseq_df = get_rnaseq(ribo_handle, name)
    except NORNASEQ as e:
        print(e)
        exit(1)

    out_str = rnaseq_df.to_csv(sep = sep, header = True)
 
    with flex_out(output) as output_stream:
        print(out_str, file = output_stream)
    
    ribo_handle.close()



@check_experiment_list_in_ribo_handle
def delete_rnaseq(ribo_handle, name):
    """
    Returns rnaseq profile of the data if it exists

    Parameters
    ----------
    ribo_handle: h5py.File
        Open ribo file handle

    name: str
        Name of the experiment

    Returns
    -------
    rnaseq_df: pd.DataFrame 
        transcript expression table.
        Indices are transcript names
    """

    if not rnaseq_exists(ribo_handle, name):
        raise NORNASEQ("{} doesn't have RNA_seq data".format(name))

    del ribo_handle[EXPERIMENTS_name][name][RNASEQ_name]


@make_cli_function
def delete_rnaseq_wrapper(ribo_file, name, force = True):
    """
    Wrapper for delete_rnaseq
    """

    prompt_message    = "This will delete existing RNA-Seq data."

    with h5py.File(ribo_file , "r+") as ribo_handle:
        attr_exists = rnaseq_exists(ribo_handle, name)
        prompt_user( message     = prompt_message, 
                     attr_exists = attr_exists, 
                     force       = force)
        delete_rnaseq(ribo_handle, name)








# -*- coding: utf-8 -*-

import h5py

from .settings import *
from .core.get_gadgets import *
from .io.metadata import get_ribo_file_metadata_for_display,\
                         metadata_dict_to_aligned_str

from .core.verify import make_cli_function
###########################################################

@make_cli_function
def ribo_file_info(ribo_file):
    """
    Returns metadata about the ribo file
    and / or a given experiment if given.

    Parameters
    ----------

    ribo_file : h5py.File
        Path of the ribo file

    """
    with h5py.File(ribo_file, "r") as ribo_handle:
        ribo_info_str = ribo_info(ribo_handle)

    return ribo_info_str



def ribo_info_base(ribo_handle):
    ribo_metadata = get_ribo_file_metadata_for_display(ribo_handle)

    display_lines = metadata_dict_to_aligned_str(ribo_metadata)

    return "Ribo File Info:\n" +\
           "---------------\n" + display_lines



def _format_experiment_info( exp_info_list ):
    """
    Formats experiment information by aligning columns

    The list exp_info_list is of the form
    [ [name, reads, rnaseq, metadata] ]
    """
    num_of_columns = 5
    col_max_lens   = []
    spacer         = "  "

    for i in range(num_of_columns):
        this_max = 0
        for entry_list in exp_info_list:
            this_length = len(entry_list[i])
            if this_length > this_max:
                this_max = this_length
        col_max_lens.append(this_max)

    formatted_list = []
    for entry_list in exp_info_list:
        raw_name  = entry_list[0]
        line_list = [ raw_name + \
                      (" " * (col_max_lens[0] - len(raw_name))) ]

        for i in range(1,num_of_columns):
            raw_val = entry_list[i]
            val     = ( " " * (col_max_lens[i] - len(raw_val)) ) + \
                        raw_val

            line_list.append(val)

        formatted_list.append(spacer.join( line_list ))

    return formatted_list


def ribo_info(ribo_handle):
    """
    Returns essential information about the ribo file.
    """

    experiment_list = get_experiment_names(ribo_handle)
    display_lines    = []
    display_base     = ribo_info_base(ribo_handle)

    experiment_str = "Experiments:\n------------" \
        if len(experiment_list) > 1 else "Library:\n--------"
    display_lines.append("\n" + experiment_str + " ")

    raw_exp_info_list = [[ "Name", "Reads", "Coverage", "RNA-Seq", "Metadata" ]]

    for experiment in experiment_list:
        raw_exp_info_list.append( get_experiment_info(ribo_handle, experiment) )


    display_lines += _format_experiment_info(raw_exp_info_list)

    return display_base + "\n" + "\n".join(display_lines)


def get_experiment_info(ribo_handle, experiment):
    """
    Returns essential information about the experiment.

    The returned value is a list of the form
    [name, reads, rnaseq, metadata]
    """
    exp_handle = ribo_handle[EXPERIMENTS_name][experiment]

    if has_coverage_data(ribo_handle, experiment):
        coverage_str = "*"
    else:
        coverage_str = ""

    if rnaseq_exists(ribo_handle, experiment):
        rnaseq_str = "*"
    else:
        rnaseq_str = ""

    if exp_handle.attrs.get(USER_METADATA, ""):
        metadata = "*"
    else:
        metadata = ""

    total_reads = get_total_reads(ribo_handle, experiment)

    return [experiment, str(total_reads), coverage_str, rnaseq_str, metadata]



from .main import *
from ..rnaseq import *

@cli.group()
def rnaseq():
    """
    Display, set or delete RNA-Seq data 
    """
    pass


@rnaseq.command()
@click.argument('ribo', type = click.Path( ))
@click.option('--name',  
               help     = "experiment name", 
               type     = click.STRING , 
               required = True)
@click.option('--rnaseq',  
               help     = "Transcript Expression File", 
               type     = click.Path(exists = True) , 
               required = True)
@click.option('--sep',  
               help     = "Column Separator: Default is tab", 
               type     = click.STRING ,
               default  = "\t",  
               required = False)
@click.option( '--force',
               is_flag = True,
               help    = 'Set RNA-Seq without prompting user.')
def set(ribo, name, rnaseq, sep, force):
    """
    Store the transcript expression data of a experiment

    RNA-Seq data is given in a text file with two columns where the
    first column has the transcript names and the second column
    has transcript expressions.

    By default the columns are tab separated. 
    The user can change this by providing
    --set argument.

    The experiment name is given in the name parameter.

    \b
    Examples:
    ----------

    1) Store rnaseq data from a zipped tab separated file for experiment WT

    ribog rnaseq set --name WT --rnaseq wt_transcript.tsv.gz test.ribo

    2) Pipe rnaseq data from another process in comma separated form

    cat transcript.csv | ribog rnaseq set --name WT --sep "," test.ribo

    """

    set_rnaseq_wrapper(ribo_file   = ribo, 
                       name        = name, 
                       rnaseq_file = rnaseq, 
                       sep         = sep,
                       force       = force)


@rnaseq.command()
@click.argument('ribo', type = click.Path( ))
@click.option('--name',  
               help     = "experiment name", 
               type     = click.STRING , 
               required = False)
@click.option('--out',  
               help     = "Output File", 
               type     = click.Path() , 
               required = False)
@click.option('--sep',  
               help     = "Column Separator: Default is tab", 
               type     = click.STRING ,
               default  = "\t",  
               required = False)
def get(ribo, name, out, sep):
    """
    Get transcript expression data of a given experiment

    If no output parameter is provided, the results are printed to standard output.

    Transcript expression is reported in two columns 
    where the first column corresponds to transcript names and
    the second column corresponds to transcript expression.

    \b
    Examples
    --------

    Save the transcript expression in a tab separated file in gzipped form.

    1) ribog rnaseq get --name WT --out transcript_exp.tsv.gz test.ribo

    Print the transcript expression on the screen

    2) ribog rnaseq get --name WT test.ribo
    """

    get_rnaseq_wrapper(ribo_file = ribo, 
                       name      = name,
                       output    = out,
                       sep       = sep)



@rnaseq.command()
@click.argument('ribo', type = click.Path( ))
@click.option('--name',  
               help     = "experiment name", 
               type     = click.STRING , 
               required = True)
@click.option( '--force',
               is_flag = True,
               help    = 'Delete RNA-Seq without prompting user.')
def delete(ribo, name, force):
    """
    Delete RNA-Seq data of a particular experiment

    \b
    Example
    -------

    ribog rnaseq delete --name WT test.ribo 
    """

    delete_rnaseq_wrapper(ribo_file = ribo, 
                          name      = name,
                          force     = force)



# -*- coding: utf-8 -*-
import unittest
from unittest import mock
from unittest.mock import patch
import os
from io import StringIO, BytesIO

import numpy as np
import h5py

from test_data import *
from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.region_counts import get_extended_boundaries,\
                                           find_region_counts

from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *

from ribogadgets.io.separate_by_length import *


###########################################

TRANSCRIPT_LENGTHS=\
"""GAPDH    90
VEGFA   12
FLT 40
MYC 1
P53 85"""


TRANSCRIPT_ANNOTATION=\
"""GAPDH    0   20  UTR5    0   -
GAPDH   20  50    CDS 0   -
GAPDH   50    90 UTR3   0   -
VEGFA   0   10 CDS 0   +
VEGFA   10 12 UTR3    0   +
FLT   0 15 UTR5    0   +
FLT   15 40 CDS    0   +
MYC 0  1    CDS 0   -
MYC 1    1    UTR3    0   -
P53 0   20  UTR5    0   +
P53 20  55  CDS 0   +
P53 55  85  UTR3    0   +"""

# Revise the reads !!!!
READ_SET_1=\
"""GAPDH 2  4000  read_UTR5_1 0   +
GAPDH 3  5000  read_UTR5_2 0   +
GAPDH 10  14000  read_UTR5_3 0   +
GAPDH 10  14000  read_UTR5_4 0   +
GAPDH 11  14000  read_UTR5_5 0   +
GAPDH 12  14000  read_UTR5_6 0   +
GAPDH 14  15  read_UTR5_7 0   +
GAPDH 15  20000  read_UTR5_junc_1 0   +
GAPDH 16  18  read_UTR5_junc_2 0   +
GAPDH 17  20000  read_UTR5_junc_3 0   +
GAPDH 20  22  read_UTR5_junc_4 0   +
GAPDH 23  24  read_UTR5_junc_5 0   +
GAPDH 24  30000  read_CDS_1 0   +
GAPDH 30  32  read_CDS_2 0   +
GAPDH 44  46  read_CDS_3 0   +
GAPDH 45  55000  read_UTR3_junc_1 0   +
GAPDH 53  57000  read_UTR3_junc_2 0   +
GAPDH 54  57000  read_UTR3_1 0   +
VEGFA 0  1000  read_UTR5_junc_1 0  +
VEGFA 3  4  read_UTR5_junc_2 0  +
VEGFA 4  7  read_CDS_1 0  +
VEGFA 4  7  read_CDS_2 0  +
VEGFA 11  12000  read_UTR3_junc_1 0  +
FLT 36  39  read_UTR3_junc_1 0 +
MYC 0  1000  read_UTR5_junc_1 0  +
P53 5  8  read_UTR5_1 0  +
P53 25  27  read_CDS_1 0  +"""

#########################################
#####  Expected EXTENDED ANNOTATION 

GAPDH_regions = ( (0, 15), (15, 24), (24, 45), (45, 54), (54, 90) )
VEGFA_regions = ( (0, 0), (0, 4), (4, 5), (5, 12), (12, 12) )
FLT_regions   = ( (0, 10), (10, 19), (19, 35), (35, 40), (40, 40) )
MYC_regions   = ( (0, 0), (0, 1), (1, 1), (1, 1), (1, 1) )
P53_regions   = ( (0, 15), (15, 24), (24, 50), (50, 59), (59, 85) ) 
extended_boundary_regions = ( GAPDH_regions, VEGFA_regions,
                              FLT_regions, MYC_regions, P53_regions )
#########################################

#########################################

### Expected Region Counts (for all read lengths)
# Order is UTR5, UTR5_junc, CDS, UTR3_junc, UTR3

GAPDH_counts_all = ( 7, 5, 3, 2, 1 )
VEGFA_counts_all = ( 0, 2, 2, 1, 0 )
FLT_counts_all   = ( 0, 0, 0, 1, 0 )
MYC_counts_all   = ( 0, 1, 0, 0, 0 )
P53_counts_all   = ( 1, 0, 1, 0, 0 )

expected_counts_all_lengths = ( GAPDH_counts_all, 
                                VEGFA_counts_all, FLT_counts_all,
                                MYC_counts_all, P53_counts_all )

GAPDH_counts_1_to_3 = ( 1, 3, 2, 0, 0 )
VEGFA_counts_1_to_3 = ( 0, 1, 2, 0, 0 )
FLT_counts_1_to_3   = ( 0, 0, 0, 1, 0 )
MYC_counts_1_to_3   = ( 0, 0, 0, 0, 0 )
P53_counts_1_to_3   = ( 1, 0, 1, 0, 0 )

expected_counts_1_to_3 = ( GAPDH_counts_1_to_3, 
                           VEGFA_counts_1_to_3, FLT_counts_1_to_3,
                           MYC_counts_1_to_3, P53_counts_1_to_3 )

GAPDH_counts_2_to_3 = ( 0, 2, 2, 0, 0 )
VEGFA_counts_2_to_3 = ( 0, 0, 2, 0, 0 )
FLT_counts_2_to_3   = ( 0, 0, 0, 1, 0 )
MYC_counts_2_to_3   = ( 0, 0, 0, 0, 0 )
P53_counts_2_to_3   = ( 1, 0, 1, 0, 0 )

expected_counts_2_to_3 = ( GAPDH_counts_2_to_3, 
                           VEGFA_counts_2_to_3, FLT_counts_2_to_3,
                           MYC_counts_2_to_3, P53_counts_2_to_3 )

#########################################


LEFT_SPAN  = 5
RIGHT_SPAN = 3

ANNOTATION = [\
               ([0, 20], [20, 50], [50, 90]),\
               ([0, 0], [0, 10], [10, 12]),\
               ([0, 15], [15, 40], [40, 40]),\
               ([0, 0], [0, 1], [1, 1]),\
               ([0, 20], [20, 55], [55, 85])
              ]


def _get_transcripts( file_in_string ):
    rows        = file_in_string.split("\n")
    pairs       = tuple( map( lambda x: x.split(), rows ) )
    ref_names   = tuple( map( lambda x: x[0], pairs ) )
    ref_lengths = tuple( map( lambda x: x[1], pairs ) )
    ref_lengths = tuple (map( int, ref_lengths ) )

    return (ref_names, ref_lengths)


class TestSeparate(unittest.TestCase):

    def setUp(self):
        self.handle = h5py.File(BytesIO())
        self.files_to_be_removed = list()

    def tearDown(self):
        self.handle.close()
        [os.remove(f) for f in self.files_to_be_removed]

    def test_separate_by_length_1_3(self):
        input_stream = StringIO(READ_SET_1)

        separated_reads = separate_by_length(input_stream, length_min = 1, length_max = 3)
        self.files_to_be_removed += separated_reads

        self.assertEqual( len(separated_reads), 3 )

        with open(separated_reads[1], "r") as sep_stream:
            length_2_lines = sep_stream.readlines()
            self.assertEqual(len(length_2_lines), 5)
            parsed_lines = [ k.strip().split() for k in length_2_lines ]
            self.assertListEqual( parsed_lines[1] , 
                     ['GAPDH', '20', '22', 'read_UTR5_junc_4', '0', '+'] )

        with open(separated_reads[0], "r") as sep_stream:
            length_1_lines = sep_stream.readlines()
            self.assertEqual(len(length_1_lines), 3)
            parsed_lines = [ k.strip().split() for k in length_1_lines ]
            self.assertListEqual( parsed_lines[0] ,
                     ['GAPDH', '14', '15', 'read_UTR5_7', '0', '+'])    


if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-

from functools import partial
from multiprocessing import Pool
import os

from ..io.separate_by_length import separate_by_length
from .quantify import quantify_experiment
from .get_gadgets import *
from ..io.write_profile import write_profile
from ..metadata import set_metadata
from ..settings import *


def create_experiment(ribo_exp_handle,
                   experiment_name, 
                   alignment_file_handle,
                   ref_names, 
                   ref_lengths, 
                   region_coordinates,
                   metagene_radius, 
                   left_span,  right_span,
                   length_min, length_max,
                   metadata        = None, 
                   store_coverage  = False,
                   nprocess        = 1,
                   tmp_file_prefix = "" ):

    """
    ??? Missing documentation ???
    """   

    # Step 1: Write Metadata
    if metadata:
        set_metadata(ribo_exp_handle, "", metadata = metadata)

    

    # Step 2: Separate files by length
    if tmp_file_prefix == "":
        tmp_file_prefix = experiment_name + "."

    files_by_length = separate_by_length(
                          input_stream = alignment_file_handle, 
                          length_max   = length_max, 
                          length_min   = length_min, 
                          file_prefix  = tmp_file_prefix)

    # files_by_length is a tuple of the form
    #   ( (file_1.bed, length_min), (file_2.bed, length_min+1), ... , 
    #     (file_k.bed, length_max) )


    quantify = partial( quantify_experiment, 
                           ref_names          = ref_names, 
                           ref_lengths        = ref_lengths, 
                           region_coordinates = region_coordinates, 
                           metagene_radius    = metagene_radius, 
                           left_span          = left_span, 
                           right_span         = right_span )

    # Step 3: Map these files to count structures ( metagene, regions counts etc )
    #quant_data_by_length = map( quantify, files_by_length )
    with Pool( nprocess ) as p:
        profiles = tuple( p.map(quantify, files_by_length) )

    for f in files_by_length:
        os.remove(f)

    # Step 4: write the result to the ribo file
    # The results are merged in the write function
    write_profile( exp_handle      = ribo_exp_handle , 
                   profiles        = profiles,
                   length_min      = length_min, 
                   length_max      = length_max,
                   left_span       = left_span,   
                   right_span      = right_span, 
                   metagene_radius = metagene_radius,
                   store_coverage  = store_coverage )
    


# -*- coding: utf-8 -*-
from collections import OrderedDict
import os
from io import IOBase

import h5py
import yaml

from .settings import *
from .io.file import get_alignment_file_handle
from .core.get_gadgets import *
from .core.exceptions import *
from .io.metadata import metadata_dict_to_aligned_str
from .io.file import (open_by_extension,
                      read_all_lines, 
                      flex_open) 
from ._version import (__format_version__,
                      __version__)
from .core.verify import prompt_user, make_cli_function ,\
                         check_experiment_list_in_ribo_handle


###########################################################

# From : https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
# This is a fix to ordering metadata
# Though it is not very important, we want to see the
# metadata in the same order.

_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG

def dict_representer(dumper, data):
    return dumper.represent_dict(data.iteritems())

def dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))

yaml.add_representer(OrderedDict, dict_representer)
yaml.add_constructor(_mapping_tag, dict_constructor)

# TODO
# Check experiment existence

###########################################################

@check_experiment_list_in_ribo_handle
def set_metadata(ribo_handle, name, metadata):
    """
    Set the user-provided metadata

    Note that this will over-write the existing metadata

    Metadata needs to be in yaml format.
    It can be a dictionary, string or a filestream.

    If no name is given, the metadata of the root ribofile
    is going to be set.

    Parameters
    ----------
    ribo_handle: h5py.File
        Open ribo file handle

    name: str
        Name of the experiment

    metadata: dict, str or IOStream
        metadata in yaml format

    Returns
    -------
    error: str
      Error returned by yaml module.
      It is the empty string if there are no errors.
    """

    handle = ribo_handle

    if name:
        handle = ribo_handle[EXPERIMENTS_name][name]

    metadata_type = type(metadata)
    # make sure the metadata is fine
    # before storing
    if metadata_type == str or issubclass(metadata_type, IOBase):
        try:
            meta_dict = yaml.safe_load(metadata)
            if not issubclass(type(meta_dict), dict):
                e_msg = " Entries must be in the form\n" +\
                        " key : value"
                raise (yaml.error.YAMLError(e_msg))
        except yaml.error.YAMLError as e:
            return "Invalid Metadata\n" + str(e)
    else:
        meta_dict = metadata

    meta_str = yaml.safe_dump(meta_dict)

    handle.attrs[USER_METADATA] = meta_str

    return ""
 
@make_cli_function
def set_metadata_wrapper(ribo_file, name, meta_file, force):
    """
    Wrapper for set_metadata
    """

    with flex_open(meta_file) as file_handle,\
          h5py.File(ribo_file, "r+") as ribo_handle:

        # <Prompt>
        handle = ribo_handle

        if name:
            handle = ribo_handle[EXPERIMENTS_name][name]

        prompt_attr    = handle.attrs.get(USER_METADATA, None)
        prompt_message = "This will over-write the existing metadata."
        prompt_user(message     = prompt_message, 
                    attr_exists = prompt_attr , 
                    force       = force)

        # </Prompt>


        error = set_metadata(ribo_handle = ribo_handle,
                             name        = name,
                             metadata    = file_handle)

        if error:
            print(" Failure:\n", error)

@check_experiment_list_in_ribo_handle
def get_metadata(ribo_handle, name):
    """
    Returns user-defined metadata in a dictionary

    If no name is given, the metadata of the root ribo file
    is returned.

    Parameters
    ----------
    ribo_handle: h5py.File
        Open ribo file handle

    name: str
        Name of the experiment

    Returns
    -------
    metadata in dictionary form    
    """

    handle = ribo_handle

    if name:
        handle = ribo_handle[EXPERIMENTS_name][name]

    metadata = handle.attrs.get(USER_METADATA, "")

    return yaml.safe_load(metadata)

@make_cli_function
def get_metadata_wrapper(ribo_file, name):
    """
    Wrapper for get_metadata
    """

    with h5py.File(ribo_file, "r") as ribo_handle:
        metadata = get_metadata(ribo_handle, name)
    
    if metadata: 
        print(yaml.safe_dump(metadata, default_flow_style = False))

    # We can not align nested dictionaries for now.
    # So we are not using fancy output
    #print( metadata_dict_to_aligned_str(metadata) )

@check_experiment_list_in_ribo_handle
def delete_metadata(ribo_handle, name):
    """
    Deletes user-provided metadata

    Parameters
    ----------
    ribo_handle: h5py.File
        Open ribo file handle

    name: str
        Name of the experiment

    """

    handle = ribo_handle

    if name:
        handle = ribo_handle[EXPERIMENTS_name][name]

    if handle.attrs.get(USER_METADATA, None):
        print("Deleting metadata")
        del handle.attrs[USER_METADATA]

    return ""

@make_cli_function
def delete_metadata_wrapper(ribo_file, name, force):
    """
    Wrapper for delete_metadata
    """

    with h5py.File(ribo_file, "r+") as ribo_handle:

        # <Prompt>
        handle = ribo_handle

        if name:
            handle = ribo_handle[EXPERIMENTS_name][name]

        prompt_attr    = handle.attrs.get(USER_METADATA, None)
        prompt_message = "Existing metadata is going to be deleted."
        prompt_user(message     = prompt_message, 
                    attr_exists = prompt_attr , 
                    force       = force)

        # </Prompt>
        delete_metadata(ribo_handle, name)
        
    



# -*- coding: utf-8 -*-
from collections import OrderedDict

import h5py
import pandas as pd
import numpy as np


def write_profile( profile_array, lengths_array, hdf5_handle, metadata ):

    start_site_coverage = merge_metagene() 
    stop_site_coverage = merge_metagene()

    # Do this in array - pyothonic way
    #UTR3_counts = merge_region_counts()

     



# -*- coding: utf-8 -*-

from .main import *
from ..create import create_ribo_file_wrapper

###########################################################

@cli.command()
@click.argument('ribofile',      type = click.Path( ))
@click.option('-a', '--alignmentfile',
              help     = "Aligned Ribo-Seq Data File\n"
                         "If no file is provided, "
                         "it is read from standard input."
                         "Alignment file can be in bed or bam format."
                         "Bed files with '.gz' or '.gzip'" 
                         "extension are assumed to be gzipped." 
                         "So they will be automatiocally unzipped."
                         "If no aligment file is given," 
                         "it is read from the standard input."
                         "Thus, the user can pipe the output of some process"
                         "to ribogadgets to produce ribo files. ",
              type     = click.Path(exists = True),
              required = False)
@click.option('-f', '--alignmentformat' ,
               help         = "Alignment File Format",
               type         = click.Choice(['bed', 'bam']),
               default      = "bed",
               show_default = True,
               required     = False)
@click.option('--name',
               help     = "Experiment name\n"
                          "If name is not provided, experiment "
                          "name is extracted from the ribo file path. "
                          "If --name parameter is omitted, experiment name is determined "
                          "using the ribo file path. For example if the ribo file path is "
                          "/home/user/data/WT.ribo, "
                          "then, experiment name is set to 'WT'. Experiment name can only contain "
                          "alphanumeric characters, '_', '-' and '.'. ",
               type     = click.STRING ,
               required = False)
@click.option('--reference',
               help     = "Reference name\n"
                          "It is good practice to give a unique "
                          "name to a particular "
                          "transcriptome and annotation pair and "
                          "use it consistently."
                          "For convenience, the user has the freedom" 
                          "to choose a name for the transcript reference "
                          "and the annotation, used to create a ribo file,"
                          "If used consistently, one can tell whether "
                          "two ribo files are coming from the same annotation or not."
                          "Note that ribo files having different reference names "
                          "can not be merged.",
               type     = click.STRING ,
               required = True)
@click.option('--lengthsfile', '--lengths',
              help     = "A tab-separated file containing"
                         " ref. name and ref. lengths",
              type     = click.Path(exists = True),
              required = True)
@click.option('--annotationfile', '--annotation',
              help     = "A bed file defining UTR5,"
                         " CDS and UTR3 regions."
                         "Each transcript must be annotated. "
                         "More explicitly, the coordinates of the regions UTR5, " 
                         "CDS and UTR3 must be provided for each transcript. "
                         "This annotation is provided in a BED file. "
                         "The annotation can not contain gaps. "
                         "All nucleotides of a transcript must belong to a region. "
                         "If a transcript does not have a CDS or UTR3 region, "
                         "it must be excluded from the lengths file and annotation. "
                         "Note that bed files are 0-based, "
                         "start position is included "
                         "and the end position is excluded. "
                         "The order and transcript names in the lengths file "
                         "must match the order in the annotation file. "
                         "Annotation data is kept in ribo files. "
                         "It is possible to extract the annotation from a ribo file "
                         "using the dump command.",
              type     = click.Path(exists = True),
              required = True)
@click.option('--metageneradius',  '--radius',
              help     = "Number of nucleotides on either side of start"
                         " / stop sites for metagene analysis.\n\n"
                         "Aligning transcripts by their start / stop sites"
                         "and aggregating the coverage gives us metagene data."
                         "The number of nucleotides to be taken on either side of"
                         "start / stop site is given by the metagene radius.",
              type     = click.INT,
              default  = 50,
              required = False)
@click.option('--leftspan', '-l',
              help         = "Number of nucleotides to the left of start"
                             " / stop sites for defining UTR5 / UTR3"
                             " junction regions.",
              type         = click.INT,
              default      = 35,
              show_default = True,
              required     = False)
@click.option('--rightspan', '-r',
              help         = "Number of nucleotides to the right of start"
                             " / stop sites for defining "
                             "UTR5 / UTR3 junction regions.",
              type         = click.INT,
              default      = 15,
              show_default = True,
              required     = False)
@click.option('--lengthmin', '--min',
              help         = "Minimum read length to be counted",
              type         = click.INT,
              default      = 15,
              show_default = True,
              required     = False)
@click.option('--lengthmax', '--max',
              help         = "Maximum read length to be counted",
              type         = click.INT,
              default      = 35,
              show_default = True,
              required     = False)
@click.option('--ribometa',
              help     = "Metadata file, for the ribo file, in yaml format.",
              type     = click.Path(exists = True),
              required = False)
@click.option('--expmeta',
              help     = "Metadata file, for the experiment, in yaml format.",
              type     = click.Path(exists = True),
              required = False)
@click.option('--rnaseq',
              help     = "Transcript expression file in tab-separated format.",
              type     = click.Path(exists = True),
              required = False)
@click.option( '--nocoverage',
               is_flag = True,
               help    = "Do not store coverage. "
                         "By default, coverage IS stored "
                         "at nucleotide resolution, "
                         "of each transcript, for each read length. "
                         "Turning this flag on decreases size of ribo file "
                         "at the cost of coverage data. "
                         "From the alignment data, five prime end of the reads, "
                         "mapping to each nucleotide position, "
                         "of each transcript is computed for each read length. "
                         "We call this coverage. "
                         "Coverage is used for metagene analysis and region counts. "
                         "By default, ribogadgets stores coverage data in the ribo file. "
                         "Keeping coverage increases the ribo file size considerably. "
                         "Users can choose NOT to keep coverage by setting this flag. ")
@click.option('--nprocess', '-n',
              help         = "Number of cores to be used.",
              type         = click.INT,
              default      = 1,
              show_default = True,
              required     = False)

def create(ribofile,
           alignmentfile,
           alignmentformat,
           lengthsfile,
           name,
           reference,
           annotationfile,
           metageneradius,
           leftspan,      rightspan,
           lengthmin,     lengthmax,
           nocoverage,
           ribometa,
           rnaseq,
           expmeta,
           nprocess):
    """Creates a ribo file from a given reference, annotation and alignment file.

    The resulting ribo file contains a single experiment
    whose name is provided in "--name". If "--name" is not provided,
    the experiment name will be extracted from the ribo file path
    after stripping the file extension.

    Ribogadgets works on transcriptomic coordinates.
    In other words, all coordinates are relative to transcripts
    where the first nucleotide of the transcript is always zero.
    Each entry in the alignment reference must come from a single transcript.
    Ribo-Seq data aligned against genomic coordinates is NOT usable
    with ribogadgets.

    A note on transcript <-> reference:
     Sequencing reads are mapped against a reference
     to generate alignment files in bed or bam format. 
     Each entry of this reference is coming
     from a transcript. Therefore, in this context, refrence and transcript
     correspond to the same entity. Thus, the names 'reference' and 'transcript'
     are used interchangebly. For example 'reference names' and
     'transcript names' refer to the same list of names.

    Reference Lengths File:
     Reference lengths must be provided in the 'lengths' option.
     This must be a tab separated file where transcript names
     are in the first coulmn and
     the transcript lengths are in the seocond column.

         \b
      
         Example:
          ============    ====
          TRANSCRIPT_1    1512
          TRANSCRIPT_2    1387
          ============    ====


    Annotation Examples:
     Below are some valid and invalid annotation file examples.
     Ribogadgets will report an error and fail to generate a ribo file
     if an invalid annotation file is given.

         \b
     
         VALID ANNOTATION:
          (Assuming length of TRANSCRIPT_1 is 1252)
           
           ============  ====  ====  =====   ==  ===
           TRANSCRIPT_1  0     21     UTR5    0   \+
           TRANSCRIPT_1  21    1041   CDS     0   \+
           TRANSCRIPT_1  1041  1252   UTR3    0   \+
           ============  ====  ====  =====   ==  ===

         \b
          
         INVALID ANNOTATION:
          (Assuming length of TRANSCRIPT_2 is 1000.
          Nucleotide positions 32,33 and 34 are not annotated)
           
           ============  ===  =====   =====  ==  ===
           TRANSCRIPT_1  0     32     UTR5    0  \+
           TRANSCRIPT_1  35    920    CDS     0  \+
           TRANSCRIPT_1  920   1000   UTR3    0  \+
           ============  ===  =====   =====  ==  ===

         \b
         
         INVALID ANNOTATION:
          (Assuming length of TRANSCRIPT_3 is 1200)
          (There is no CDS region.)
     
           ============  ==  ======   =====  ==  ==
           TRANSCRIPT_3  0     50     UTR5    0  \+
           TRANSCRIPT_3  50    1200   UTR3    0  \+
           ============  ==  ======   =====  ==  ==

         \b
         
         INVALID ANNOTATION:
          (Assuming length of TRANSCRIPT_4 is 1000)
          (There is no UTR3 region.)
     
           ============  ===  ======   =====  ==  ==
           TRANSCRIPT_3  0     700     UTR5    0  \+
           TRANSCRIPT_3  100    1000   CDS     0  \+
           ============  ===  ======   =====  ==  ==

    Region Counts & Left / Right Span:
     For each transcript, the number of reads aligning to
     each region (UTR5, CDS, and UTR3) are stored.
     For this quantification, we exclude nucleotides in some
     proximity of start and stop sites.

     This proximity is defined by taking
     'leftspan' many nucleotides to the left of start / stop sites and
     'rightspan' many nucleotides to the right of start / stop sites.
     The regions around start and stop sites are called
     UTR5_junction and UTR3_junction respectively.
      \b
      
       UTR5_juction: Nucleotides 'around' start site.
                     This region is between UTR5 and CDS.
      \b
      
       UTR3_juction: Nucleotides 'around' stop site.
                     This region is between CDS and UTR3.
                     
       Note that 'around' is precisely defined by 'leftspan' and 'rightspan'
       arguments.


    Read Length Range:
     Quantified Ribo-Seq data (metagene coverage, region counts,
     transcript ocoverage (if any))
     is stored for each read length for a given range.
     This range is defined by --lengthmin and --lengthmax.
     Both values are inclusive.
     So, for example, for --lengthmin 19 --lengthmax 21,
     region counts are computed and stored for
     RNA fragments of length 19,20 and 21 separetely.
     For human Ribo-Seq data, a range from 15 to 35
     can be sufficient for conventional experiments.

    Metadata:
     Metadata, either for the ribo file or for any experiment,
     is an optional argument. Metadata must be provided in yaml format.

     \b
     Users can provide metadata for
        i)  ribo file
           --ribometa
        ii) experiment
           --expmeta
    
     Metadata is provided in pairs of the form 'label: value'

        \b
                
        Example Metadata:
         ============  ======
          cell-line:   HEK
          enzyme:      RNASEI
         ============  ======

    Transcript Expression (RNA-Seq) Data:
     ?? This part need reviosion as we are changing RNA-Seq format ??
     RNA-Seq data is an optional part of ribo file.
     If there is RNA-Seq data to be paired with the Ribo-Seq experimnet data,
     it can be provided using the --rnaseq option.
     RNA-Seq data must be provided in a tab separated file
     where the first column has the transcript names and the
     second column has the transcript abundance.

        \b
                    
        Example RNA-Seq File:
          ============   =====
          TRANSCRIPT_1    42
          TRANSCRIPT_2    3
          TRANSCRIPT_3    15.2
          ============   =====


    Examples:
     Below are some ribo file generation examples in different settings.
     
     1) Some command line exaples to create ribo files.
     Create a file named WT.ribo in the current directory.
     Provide alignment data in a zipped bed file.
     Do NOT store coverage data.

            \b 
            
            .. code:: bash      
             
             ribog create --name WT \\
                --alignmentfile WT.bed.gz \\
                --reference appris_human_v1 \\
                --lengths appris_len.tsv \\
                --annotation appris_regions.bed \\
                --radius 50 \\
                -l 35 -r 15 \\
                --lengthmin 15 --lengthmax 35 \\
                --nocoverage \\
                WT.ribo     
             
             
 

     2) Create a file named Treatment_1.ribo in the current directory.
     Provide alignment data in a zipped bam file.
     Store coverage data.

     \b
     
     .. code:: bash
     
             ribog create --name Treatment_1 \\
                 --alignmentfile Treatment_1.bam \\
                 --format bam \\
                 --reference appris_human_v1 \\
                 --lengths appris_len.tsv \\
                 --annotation appris_regions.bed \\
                 --radius 50 \\
                 -l 35 -r 15 \\
                 --lengthmin 15 --lengthmax 35 \\
                Treatment_1.ribo
                
     3) Create a file named WT_2.ribo in the current directory.
     Provide alignment data in a zipped bed file.
     Do NOT store coverage data.
     Read metadata of this experiment from WT_2_meta.yaml

     \b
                 
     .. code:: bash 
     
             ribog create --name WT_2 \\
                 --alignmentfile WT_2.bed.gz \\
                 --reference appris_human_v1 \\
                 --lengths appris_len.tsv \\
                 --annotation appris_regions.bed \\
                 --radius 50 \\
                 -l 35 -r 15 \\
                 --lengthmin 15 --lengthmax 35 \\
                 --nocoverage \\
                 --expmeta WT_2_meta.yaml \\
                 WT_2.ribo  
    """

    print("creating the ribo file {}...".format(ribofile))

    create_ribo_file_wrapper(
                     ribo_file_path      = ribofile,
                     experiment_name     = name,
                     alignment_file      = alignmentfile,
                     alignment_format    = alignmentformat,
                     reference_name      = reference,
                     lengths_file        = lengthsfile,
                     annotation_file     = annotationfile,
                     metagene_radius     = metageneradius,
                     left_span           = leftspan,
                     right_span          = rightspan,
                     length_min          = lengthmin,
                     length_max          = lengthmax,
                     ribo_metadata       = ribometa,
                     experiment_metadata = expmeta,
                     store_coverage      = not nocoverage,
                     rnaseq_file         = rnaseq,
                     nprocess            = nprocess)

    print("Done.")



# -*- coding: utf-8 -*-
from collections import OrderedDict
import time
import os
import re

import h5py
import pandas as pd
import numpy as np

from .settings import *
from .rnaseq import set_rnaseq
from .io.file import get_alignment_file_handle
from .core.get_gadgets import *
from .core.metagene import *
from .metadata import set_metadata
from .core.exceptions import *
from .core.create_experiment import create_experiment
from .core.verify import make_cli_function
from .io.file import (open_by_extension,
                      read_all_lines, 
                      flex_open) 
from ._version import __format_version__,\
                      __version__ 

########################################################


def _get_experiment_name_from_filepath(filepath):
    """
    Extracts name from the ribo file path
    after stripping the file extension
    """

    basename        = os.path.basename(filepath)
    basename_pieces = basename.split(".")
    experiment_name = ".".join(basename_pieces[:-1])
    
    return experiment_name

def _check_experiment_name(name):
    """
    Raises an exception if the name is invalid.
    """
    
    error_message = \
    """
    Invalid Experiment Name! 
    Name must contain at least one alphanumeric character and
    it can only contain alphanumeric characters, '_', '-' and '.'.
    """
    
    alphanum_test   = re.match("[a-zA-Z0-9]+", name)
    valid_char_test = re.fullmatch("[a-zA-Z0-9\_\-.]+", name)
    
    if alphanum_test and valid_char_test:
        return True
    else:
        raise InvalidName(error_message)


def initialize(h5_handle, experiment_name, reference_name):
    """
    Basic initialization of a ribo file.
    It creates the hdf5 groups for experiments
    and reference.  

    Parameters
    ----------
    h5_handle: h5py.File
        Handle for writable hdf5 file

    experiment_name : str 
        A file group under
        experiments is created with this name

    reference_name : str 
        User defined name of the reference

    Returns
    -------
    None

    """

    _check_experiment_name(experiment_name)
    
    h5_handle.create_group(EXPERIMENTS_name)
    h5_handle.create_group(REFERENCE_name)
    h5_handle[REFERENCE_name].attrs[ATTRS_REFERENCE]      = reference_name
    h5_handle[REFERENCE_name].attrs[ATTRS_FORMAT_VERSION] = __format_version__
    h5_handle[REFERENCE_name].attrs[ATTRS_VERSION]        = __version__

    exp_group = h5_handle[EXPERIMENTS_name].create_group(experiment_name)



def set_reference_names_and_lengths(h5_handle, lengths_file):
    """
    Creates reference names and lengths fields in the references part
    of the ribo file.

    The order (of transcripts) is crucial because 
    the same order will be kept in all parts of the file.  


    Parameters
    ----------

    h5_handle: h5py.File
        Handle for writable hdf5 file

    lengths_file : str 
        Tab separated file of two columns
        First column: Transcript names
        Second Column: Transcript lengths
        One line per transcript.

    Returns
    -------
    Reference names and lengths
    """

    reference_handle = h5_handle[REFERENCE_name]
    reference_df     = pd.read_csv(lengths_file, sep = "\s+", 
                               names = ["name","length"])
    name_lengths     = list(map(len, reference_df["name"]) )
    max_len          = max(name_lengths)
    dt               = np.dtype("S" + str(max_len))
    ref_names        = np.array( reference_df["name"], dtype = dt )
    ref_lengths      = np.array( reference_df["length"], 
                                 dtype = np.uint32 )

    reference_handle.create_dataset( 
        REF_DG_REFERENCE_NAMES, 
        data        = ref_names, 
        dtype       = dt, 
        compression = DEFAULT_COMPRESSION, 
        fletcher32  = DEFAULT_FLETCHER32 )

    reference_handle.create_dataset( 
        REF_DG_REFERENCE_LENGTHS, 
        data        = ref_lengths, 
        dtype       = np.uint32, 
        compression = DEFAULT_COMPRESSION, 
        fletcher32  = DEFAULT_FLETCHER32 )
    
    return ( get_reference_names(h5_handle) , 
             get_reference_lengths(h5_handle))

#########################################################

def post_check_annotation(h5_handle, pre_annotation):
    """
    More sanity checks for the annotation.
    See the function "check_annotation" for details.
    """ 

    error_base     = "Annotation error:\n"
    error_message  = ""
    ref_lengths    = get_reference_lengths(h5_handle)
    ref_names      = get_reference_names(h5_handle)
    len_dict       = dict( tuple(zip( ref_names, ref_lengths )) )

    for this_name, boundaries in pre_annotation.items():
        utr5, cds , utr3 = tuple(boundaries.values())

        if cds[0] == cds[1]:
            error_message +=  this_name + \
                              " : must have a positive length CDS\n"

        if utr3[0] == utr3[1]:
            error_message +=  this_name + \
                              " : must have a positive length UTR3\n"

        if utr5[1] != cds[0]:
            error_message += this_name + \
                " : has gaps or overlaps in annotation. Check its boundaries.\n"

        if cds[1] != utr3[0]:
            error_message += this_name + \
                " : has gaps or overlaps in annotation. Check its boundaries.\n"

        if utr3[1] != len_dict[this_name]:
            error_message += this_name + \
                " : utr3 3 prime (right) boundary " + \
                "must be equal to the gene length.\n"

        if error_message:
            raise AnnotationError(error_base + error_message)





def check_annotation( h5_handle, annotation_lines ):
    """
    Performs some sanity checks in the given annotation file.

    Note that the transcript (reference) names and lengths
    are provided in a tsv file and recorded prior to this
    step.

    Every trasnscript must have an annotation
    and a non-zero length of CDS and UTR3 regions.

    Also there shouldn't be any gaps in the annotation.
    In other words, there shouldn't be a region of a
    transcript that bnelongs to none of CDS, UTR5 or UTR3.

    End position of one region must be equal to the start
    position of the next region. Note that the region
    boundaries are in bed form. This means, it is 0-based
    and start position is inclusive and the end position is exclusive.

    It raises an AnnotationError once it detects an inconsistency
    for a transcript.

    Parameters
    ----------

    h5_handle:  h5py.File 
        Handle to an open, readable/ writable hdf5 file

    annotation_lines: list
        An array of bed file entries

    """
    ref_lengths = get_reference_lengths(h5_handle)
    ref_names   = get_reference_names(h5_handle)

    line_number   = 0
    error_message = "Annotation error in line number: "

    recorded_annotations = OrderedDict()
    for r in ref_names:
        recorded_annotations[r] = \
            OrderedDict( ( (UTR5_name, [0,0]), 
                         (CDS_name, [0, 0]), 
                         (UTR3_name, [0,0]) ) )


    for this_line in annotation_lines:
        line_number +=1
        contents     = this_line.strip().split()
        if len(contents) < 6:
            continue

        current_ref = contents[0]
        start       = int(contents[1])
        end         = int(contents[2])
        region      = contents[3].upper()
        strand      = contents[5]

        if region not in REGION_names:
            error_message += str(line_number) + "\n"
            error_message += "\nInvalid region identifier:\n"
            error_message += "{} not in {}".format(region, REGION_names)
            raise AnnotationError(error_message)

        if start > end:
            error_message += str(line_number) + "\n"
            error_message += "\nInvalid region boundary:\n"
            error_message += "{} > {}".format(start, end)
            raise AnnotationError(error_message)

        if current_ref not in ref_names:
            error_message += str(line_number) + "\n"
            error_message += "Invalid reference name {}".\
                                format(current_ref)
            raise AnnotationError(error_message)


        recorded_annotations[current_ref][region] = [start, end]

    post_check_annotation( h5_handle, recorded_annotations)



def set_annotation( h5_handle, annotation_lines ):
    """
    Note that everything is in the transcriptomic coordinates
    and everything is with respect to + strand

    In the annotation we store a matrix of dimensions N * 3
    Each row is for one reference (in the same order as
    reference names and lengths)
    The entries are coming from BED coordinates
    <UTR5 NED POS> , <CDS END POS>, <UTR3 END POSITION>
    Since UTR5 begins with 0, and the end of one region is
    the start of the next one, the above numbers give us all
    region boundaries. 

    Parameters
    ----------

    h5_handle: h5py.File
        Handle to an open, readable/ writable hdf5 file

    annotation_lines: list 
        An array of bed file entries

    Returns
    -------

    boundaries: list( tuple )

                An array of triplets. Each triplet corresponds
                to the UTR5, CDS and UTR3 boundaries of a transcript.
                The order of the transcripts are the same as
                their order in get_reference_names(ribo)

                Thus each triplet is of the form
                ( (UTR5_left, UTR5_right), (CDS_left, CDS_right), 
                  (UTR3_left, UTR3_right) )

                These coordinates are in bed conventions.
                Therefore the left coordinates ARE inclusive 
                and the right coordinates are NOT.
                The coordinates are 0-based.

                Also,

                UTR5_right = CDS_left
                CDS_right  = UTR3_left
                UTR3_right = ref_length

                must hold.
    """

    # Raises an AnnotationError exception if 
    # there is a porblem with the annotation 
    check_annotation(h5_handle, annotation_lines)

    number_of_refs = get_number_of_references(h5_handle)
    region_dict    = OrderedDict()

    # Initialize all regions to zeros
    for region in REGION_names:
        region_dict[region] = np.full(shape      = (number_of_refs), 
                                      fill_value = 0, 
                                      dtype      = np.uint32)

    ref_index   = 0
    ref_list    = get_reference_names(h5_handle)
    current_ref = ref_list[0]
    prev_ref    = current_ref


    for this_line in annotation_lines:
        contents = this_line.strip().split()
        if len(contents) < 6:
            continue
        current_ref = contents[0]
        start       = int(contents[1])
        end         = int(contents[2])
        region      = contents[3].upper()
        strand      = contents[5]

        # increment the index if we see a new reference
        # make sure that the next ref in the bed file is the same as the\
        # ref in the ref_list
        # this also ensures that the bed file is in the correct order.

        if current_ref != prev_ref:
            ref_index +=1
            assert current_ref == ref_list[ref_index]
            prev_ref = current_ref

        region_dict[region][ref_index] = (end)


    concatenated_arrays = np.concatenate( tuple(region_dict.values()) )
    ref_array           = np.reshape(concatenated_arrays, 
                                      newshape = (number_of_refs, 3), 
                                      order    = "F" )

    h5_handle[REFERENCE_name].create_dataset(
                                REF_ANNOTATION_NAME,
                                data        = ref_array,
                                dtype       = np.uint32,
                                compression = DEFAULT_COMPRESSION, 
                                fletcher32  = DEFAULT_FLETCHER32)

    boundaries = tuple(map( lambda x: 
                              [(0, x[0]) , (x[0], x[1]), (x[1], x[2])] , 
                            ref_array ) )

    return boundaries
  
##########################################################

###########################################################    

def set_coverage_vectors_individual(pivots, radius):
    """
    Computes the total contribution of all transcripts
    around start sites.
    The pivots iterate over transcripts.
    Note that it may well happen that  some transcripts
    are short on the 5' or 3' ends so that
    they may not contribute to all radius many nucleotides on
    any end of the start / stop site.
    So we initialize contribution to 0, and increement the
    parts that are covered by the transcript.
    
    Example:
    If our transcript T is as follows:
    (Bed format coordinates)
    UTR5: 0 2
    CDS: 2 10
    UTR3: 10 15

    if radius is 3,
    its corresponding contribution vector will be
    w = (0,1,1,1,1,1,1)
    Note that the left-most contribution is 0
    because this transcript has only 2 nucleotides on UTR5

    This vector 2 will be added to result, 
    along with contribution of other transcripts,
    to compute the overal contribution around the start site.  

    """
    result = np.zeros( 2*radius + 1, dtype = np.uint32 )

    for p in tuple(pivots):
        this_coverage = np.zeros( 2*radius + 1, dtype = np.uint32 )
        start_index   = max( radius - p[1] , 0)
        end_index     = radius + 1 + min( radius , p[2] - p[1] )

        this_coverage[start_index:end_index] = 1
        result += this_coverage

    return result

###########################################################    

def set_coverage_vectors(ribo, radius):
    """
    Coverage vectors hold the contribution of the transcripts
    around start / stop sites .
    These vectors can be used to normalize metagene data
    ( start / stop sites) later. 
    """

    boundaries        = get_region_boundaries(ribo)
    start_site_pivots = map( lambda x: ( 0, x[1][0], x[2][1] ) , 
                               boundaries )
    stop_site_pivots  = map( lambda x: ( 0, x[1][1], x[2][1] ) , 
                               boundaries )

    start_site_boundaries = \
       set_coverage_vectors_individual(start_site_pivots, radius)
    stop_site_boundaries = \
       set_coverage_vectors_individual(stop_site_pivots, radius)

    ribo[REFERENCE_name].create_dataset(
                           REF_DG_START_SITE_COV, 
                           data       = start_site_boundaries, 
                           dtype      = np.uint32,
                           fletcher32 = DEFAULT_FLETCHER32 )
    ribo[REFERENCE_name].create_dataset(
                           REF_DG_STOP_SITE_COV, 
                           data       = stop_site_boundaries, 
                           dtype      = np.uint32,
                           fletcher32 = DEFAULT_FLETCHER32 )

###########################################################

def set_ribo_info(ribo, 
                  length_min, length_max, 
                  reference_name,
                  metagene_radius, 
                  left_span, right_span):

    """
    Essential ribo metadata is stored 
    in the atrributes of the hdf5 file.
    These attributes are crucial and shouldn't be changed
    after file creation.
    """

    ribo.attrs[ATTRS_VERSION]         = __version__
    ribo.attrs[ATTRS_FORMAT_VERSION]  = __format_version__

    ribo.attrs[ATTRS_LENGTH_MIN]      = length_min
    ribo.attrs[ATTRS_LENGTH_MAX]      = length_max
    ribo.attrs[ATTRS_TIME]            = time.time()
    ribo.attrs[ATTRS_METAGENE_RADIUS] = metagene_radius
    ribo.attrs[ATTRS_LEFT_SPAN]       = left_span
    ribo.attrs[ATTRS_RIGHT_SPAN]      = right_span
    ribo.attrs[ATTRS_REFERENCE]       = reference_name

##########################################################

def _get_metadata(file_path):
    if not file_path:
        return ""
    else:
        with flex_open(file_path) as metastream:
            meta_str = metastream.read()
        return meta_str

##########################################################    

def create_ribo(ribo, 
                experiment_name, 
                alignment_file,
                reference_name,
                lengths_file, 
                annotation_file,
                metagene_radius, 
                left_span, right_span,
                length_min, length_max,
                alignment_format    = 'bed',
                store_coverage      = False,
                ribo_metadata       = None,
                experiment_metadata = None,
                rnaseq_file         = None,
                nprocess            = 1,
                tmp_file_prefix     = ""):
    """
    Creates a ribo file for a single experiment.


    The essential input data of this function are
       ribo file handle, annotation, transcript length file,
       alignment file

    The essential input, as single value parameters, are:
       experiment name and reference name,
       metagene_radius, left_span, right_span,
       length_min, length_max.    

    It writes annotation, metadata and 
    ribosome profiling data of the experiment
    to the given ribo file handle.

    Parameters
    ----------
    ribo :  h5py.File 
       Handle for the open hdf5 file

    experiment_name : str
               The name of the single experiment in the ribo file.
               An datagroup under experiments will be created accordingly 
    
    lengths_file : str
               File path. 
               A tab separated file containing the reference lengths
               The first column holds the  reference names
               The second column holds the reference lengths
               Note that the reference names 
               should be exactly the same as in
               the alignment file and the annotation file
               in the SAME ORDER

    alignment_file : str
                 File path.
                 Can be a bed, bam or sam file.
                 It can also be an StringIO stream.
                 If this is an empty string,
                 the input will be read from the standard io


    reference_name : str
             User defined name for the reference set:
             (transcript sequence, annotation)

    annotation_file : str
                  File path.
                  A bed file annotation the UTR5, CDS and UTR3
                  regions of the transcirpts. 
                  This can also be a StringIO object

    metagene_radius : int
                  The number of nucleotides to the left and right of
                  the start / stop sites in the metagene analysis.

    left_span : int
            The number of nucleotides to the left of start / stop sites
            for defining UTR5 & UTR3 junction regions.

    right_span : int
             The number of nucleotides to the right of start / stop sites
             for defining UTR5 & UTR3 junction regions.

    length_min : int
             Minumum read length to be counted in the ribosome profiling.
             Inclusive.

    length_max : int
             Maximum read length to be counted in the ribosome profiling.
             Inclusive. Note that all read lengths from 
             length_min to lenght max are counted.

    store_coverage: Boolean
        Store the coverage, at length level, for all transcripts. 

    ribo_metadata : str, dict or IOStream
        User-provided metadata of the ribo file in yaml format.

    experiment_metadata : str, dict or IOStream
        User-provided metadata of the experiment in yaml format.

    rnaseq_file : str or IOStream 
       A tab separated file of two columns
       The first column corresponds to transcript names
       The second column corresponds to transcript expression

    nprocess : int
               Number of parallel processes for 
               extracting the ribosome profiling data.

    tmp_file_prefix : int
                      File name prefix for the intermediate files
                      created (and later deleted).
                      For example, reads of the same length 
                      are collected in one inetermediate file. 

    Returns
    -------
    None

    """

    initialize(ribo, experiment_name, reference_name)

    (ref_names, ref_lengths) = \
       set_reference_names_and_lengths(ribo , lengths_file)

    annotation = set_annotation( ribo, 
                                 read_all_lines(annotation_file) )

    set_coverage_vectors(ribo, metagene_radius)

    set_ribo_info(ribo, 
                  length_min, length_max,
                  reference_name, 
                  metagene_radius, 
                  left_span, right_span)

    alignment_file_handle = \
            get_alignment_file_handle(alignment_file, 
                                      alignment_format = alignment_format )

    set_metadata(ribo_handle = ribo, 
                 name        = "", 
                 metadata    = ribo_metadata)

    create_experiment(
        ribo_exp_handle       = ribo[EXPERIMENTS_name][experiment_name], 
        experiment_name       = experiment_name, 
        alignment_file_handle = alignment_file_handle,
        metagene_radius       = metagene_radius,
        ref_names             = ref_names,
        ref_lengths           = ref_lengths,
        region_coordinates    = annotation ,
        left_span             = left_span, 
        right_span            = right_span,
        length_min            = length_min, 
        length_max            = length_max,
        store_coverage        = store_coverage,
        metadata              = experiment_metadata, 
        nprocess              = nprocess,
        tmp_file_prefix       = tmp_file_prefix)

    if rnaseq_file:
        with flex_open(rnaseq_file, "rt") as input_stream:
            rnaseq_df = pd.read_csv(input_stream, 
                                    header    = None, 
                                    names     = [0], 
                                    index_col = 0, 
                                    sep       = "\t" ) 

        set_rnaseq(ribo, 
                   name      = experiment_name, 
                   rnaseq_df = rnaseq_df)



############################################################################

def create_ribo_file(ribo_file_path, experiment_name, 
                     *argv, **kwargs):
    """
    Create a ribo file in the given file path.
    This is a wrapper for the create_ribo function to be used by cli.
    See create_ribo function for details.

    Parameters
    ----------

    ribo_file_path: str or path
        The path of the ribo file

    experiment_name: str
        The name of the experiment in the ribo file

    *argv, **kwargs: passed on to create_ribo function

    """

    kwargs["ribo_metadata"]       = \
         _get_metadata(kwargs.get("ribo_metadata"))
    kwargs["experiment_metadata"] = \
         _get_metadata(kwargs.get("experiment_metadata"))
    tmp_file_prefix = ribo_file_path + "."

    if not experiment_name:
        experiment_name = _get_experiment_name_from_filepath(ribo_file_path)


    with h5py.File( ribo_file_path, "w" ) as ribo_handle:
        try:
            create_ribo(ribo_handle,
                        experiment_name,  
                        *argv, **kwargs, 
                        tmp_file_prefix = tmp_file_prefix)
        except RiboBaseError as e:
            ribo_handle.close()
            os.remove(ribo_file_path)
            raise RiboBaseError(e)
        except:
            print("An unexpected error happend!")
            ribo_handle.close()
            os.remove(ribo_file_path)
            raise
            

############################################################################

@make_cli_function
def create_ribo_file_wrapper(*argv, **kwargs):
    """
    Wrapper function for create_ribo_file
    to be used by CLI
    """

    try:
        create_ribo_file(*argv, **kwargs)
    except RiboBaseError as e:
        print("Ribo file creation has failed!")
        print(e)
        return 1

    return 0


# -*- coding: utf-8 -*-

import numpy as np

###########################################
REF_LEN_FILE     = "ref_lengths.tsv"
TEST_RIBO        = "test_create.ribo"
TEST_RIBO_2      = "test_2.ribo"
ANNOTATION_FILE  = "annotation.bed"
ALIGNMENT_FILE_1 = "alignment_file_1.bed"
ALIGNMENT_FILE_2 = "alignment_file_2.bed"
METADATA_FILE_1  = "exp_metadata.yaml"
RIBO_META_FILE_1 = "ribo_metadata.yaml"
RNASEQ_FILE_1    = "rnaseq_1.tsv"
RNASEQ_FILE_2    = "rnaseq_2.tsv"



#TRANSCRIPT_LENGTHS=\
"""GAPDH    20
VEGFA   22
MYC 17"""

TRANSCRIPT_LENGTHS=\
"GAPDH\t20\nVEGFA\t22\nMYC\t17"


TRANSCRIPT_ANNOTATION=\
"""GAPDH    0   5  UTR5    0   +
GAPDH   5  15    CDS 0   +
GAPDH   15  20    UTR3 0   +
VEGFA    0   4  UTR5    0   +
VEGFA   4  16    CDS 0   +
VEGFA   16  22    UTR3 0   +
MYC    0   3  UTR5    0   +
MYC   3  10    CDS 0   +
MYC   10  17    UTR3 0   +"""


READ_SET_1=\
"""GAPDH 1  3  len_2_UTR5_1 0   +
GAPDH 3  5  len_2_UTR5_junc_1 0   +
GAPDH 4  6  len_2_UTR5_junc_2 0   +
GAPDH 4  6  len_2_UTR5_junc_3 0   +
GAPDH 6  8  len_2_UTR5_junc_4 0   +
GAPDH 7  9  len_2_UTR5_junc_5 0   +
GAPDH 9  11 len_2_CDS_1 0   +
GAPDH 10  12  len_2_CDS_2 0   +
GAPDH 13  15  len_2_UTR3_junc_1 0   +
GAPDH 13  15  len_2_UTR3_junc_2 0   +
GAPDH 14  16  len_2_UTR3_junc_3 0   +
GAPDH 15  17  len_2_UTR3_junc_4 0   +
GAPDH 18  20  len_2_UTR3_1 0   +
GAPDH 19  21  len_2_UTR3_1 0   +
GAPDH 0  3  len_3_UTR5_junc_1 0   +
GAPDH 1  4  len_3_UTR5_junc_2 0   +
GAPDH 3  6  len_3_UTR5_1 0   +
GAPDH 3  6  len_3_UTR5_2 0   +
GAPDH 3  6  len_3_UTR5_3 0   +
GAPDH 4  7  len_3_UTR5_4 0   +
GAPDH 4  7  len_3_UTR5_5 0   +
GAPDH 5  8  len_3_UTR5_6 0   +
GAPDH 6  9  len_3_UTR5_7 0   +
GAPDH 8  11  len_3_CDS_1 0   +
GAPDH 8  11  len_3_CDS_2 0   +
GAPDH 10  13  len_3_CDS_3 0   +
GAPDH 13  16  len_3_UTR3_junc_1 0   +
GAPDH 13  16  len_3_UTR3_junc_2 0   +
GAPDH 13  16  len_3_UTR3_junc_3 0   +
GAPDH 13  16  len_3_UTR3_junc_4 0   +
GAPDH 14  17  len_3_UTR3_junc_5 0   +
GAPDH 14  17  len_3_UTR3_junc_6 0   +
GAPDH 15  18  len_3_UTR3_junc_7 0   +
GAPDH 15  18  len_3_UTR3_junc_8 0   +
GAPDH 16  19  len_3_UTR3_junc_9 0   +
GAPDH 3  7  len_4_UTR5_junc_1 0   +
GAPDH 3  7  len_4_UTR5_junc_2 0   +
GAPDH 4  8  len_4_UTR5_junc_3 0   +
GAPDH 4  8  len_4_UTR5_junc_4 0   +
GAPDH 4  8  len_4_UTR5_junc_5 0   +
GAPDH 5  9  len_4_UTR5_junc_6 0   +
GAPDH 5  9  len_4_UTR5_junc_7 0   +
GAPDH 5  9  len_4_UTR5_junc_8 0   +
GAPDH 5  9  len_4_UTR5_junc_9 0   +
GAPDH 7  11  len_4_UTR5_junc_10 0   +
GAPDH 9  13  len_4_CDS_1 0   +
GAPDH 9  13  len_4_CDS_2 0   +
GAPDH 10  14  len_4_CDS_3 0   +
GAPDH 11  15  len_4_CDS_4 0   +
GAPDH 11  15  len_4_CDS_5 0   +
GAPDH 11  15  len_4_CDS_6 0   +
GAPDH 13  17  len_4_UTR3_junc_1 0   +
GAPDH 13  17  len_4_UTR3_junc_2 0   +
GAPDH 13  17  len_4_UTR3_junc_3 0   +
GAPDH 13  17  len_4_UTR3_junc_4 0   +
GAPDH 14  18  len_4_UTR3_junc_5 0   +
GAPDH 14  18  len_4_UTR3_junc_6 0   +
GAPDH 15  19  len_4_UTR3_junc_7 0   +
GAPDH 15  19  len_4_UTR3_junc_8 0   +
GAPDH 16  20  len_4_UTR3_junc_9 0   +
GAPDH 1  6  len_5_UTR5_1 0   +
GAPDH 3  8  len_5_UTR5_junc_1 0   +
GAPDH 5  10  len_5_UTR5_junc_1 0   +
GAPDH 8  13  len_5_CDS_ 1   +
GAPDH 10  15  len_5_CDS_2 0   +
GAPDH 14  19  len_5_UTR5_junc_1 0   +
VEGFA 3  5  len_2_UTR5_junc_1 0   +
VEGFA 3  5  len_2_UTR5_junc_2 0   +
VEGFA 4  6  len_2_UTR5_junc_3 0   +
VEGFA 10  12  len_2_CDS_1 0   +
VEGFA 16  18  len_2_UTR3_junc_1 0   +
VEGFA 3  6  len_3_UTR5_junc_1 0   +
VEGFA 3  6  len_3_UTR5_junc_2 0   +
VEGFA 3  6  len_3_UTR5_junc_3 0   +
VEGFA 4  7  len_3_UTR5_junc_4 0   +
VEGFA 4  7  len_3_UTR5_junc_5 0   +
VEGFA 5  8  len_3_UTR5_junc_6 0   +
VEGFA 10  13  len_3_CDS_1 0   +
VEGFA 10  13  len_3_CDS_2 0   +
VEGFA 16  19  len_3_UTR3_junc_1 0   +
VEGFA 16  19  len_3_UTR3_junc_2 0   +
VEGFA 3  7  len_4_UTR5_junc_1 0   +
VEGFA 3  7  len_4_UTR5_junc_2 0   +
VEGFA 3  7  len_4_UTR5_junc_3 0   +
VEGFA 3  7  len_4_UTR5_junc_4 0   +
VEGFA 4  8  len_4_UTR5_junc_5 0   +
VEGFA 4  8  len_4_UTR5_junc_6 0   +
VEGFA 4  8  len_4_UTR5_junc_7 0   +
VEGFA 5  9  len_4_UTR5_junc_8 0   +
VEGFA 5  9  len_4_UTR5_junc_9 0   +
VEGFA 10  14  len_4_CDS_1 0   +
VEGFA 10  14  len_4_CDS_2 0   +
VEGFA 10  14  len_4_CDS_3 0   +
VEGFA 3  8  len_5_UTR5_junc_1 0   +
VEGFA 10  15  len_5_CDS_1 0   +
VEGFA 10  15  len_5_CDS_2 0   +
VEGFA 10  15  len_5_CDS_3 0   +
VEGFA 10  15  len_5_CDS_4 0   +
VEGFA 16  21  len_5_UTR3_junc_1 0   +
VEGFA 16  21  len_5_UTR3_junc_2 0   +
MYC 10  12  len_2_UTR3_junc_1 0   +
MYC 10  12  len_2_UTR3_junc_2 0   +
MYC 0  3  len_3_UTR5_junc_1 0   +
MYC 6  9  len_3_CDS_1 0   +
MYC 10  13  len_3_UTR3_junc_1 0   +
MYC 6  10  len_4_CDS_1 0   +
MYC 6  10  len_4_CDS_2 0   +
MYC 10  14  len_4_UTR3_junc_1 0   +
MYC 10  14  len_4_UTR3_junc_2 0   +
MYC 10  14  len_4_UTR3_junc_3 0   +
MYC 0  5  len_5_UTR5_junc_1 0   +
MYC 10  15  len_UTR3_junc_1 0   +
MYC 10  15  len_UTR3_junc_2 0   +
MYC 10  15  len_UTR3_junc_3 0   +
MYC 10  15  len_UTR3_junc_4 0   +
MYC 10  15  len_UTR3_junc_5 0   +
MYC 10  15  len_UTR3_junc_6 0   +
MYC 10  15  len_UTR3_junc_7 0   +"""


#########################################


#########################################

### Expected Region Counts for READ_SET_1
# Order is UTR5, UTR5_junc, CDS, UTR3_junc, UTR3

## LENGTH 2

GAPDH_counts_length_2 = ( 1, 5, 2, 4, 2 )
VEGFA_counts_length_2 = ( 0, 3, 1, 1, 0 )
MYC_counts_length_2   = ( 0, 0, 0, 2, 0 )

expected_counts_length_2 = np.array( (GAPDH_counts_length_2, VEGFA_counts_length_2, 
                             MYC_counts_length_2) )

ACTUAL_START_SITE_COVERAGE_length_2 = np.array( [
                                [1, 2, 0, 1, 1], 
                                [0, 2, 1, 0, 0],
                                [0, 0, 0, 0, 0]] )

ACTUAL_STOP_SITE_COVERAGE_length_2 = np.array( [ 
                                [2, 1, 1, 0, 0],
                                [0, 0, 1, 0, 0],
                                [0, 0, 2, 0, 0]] )

#########################################

## LENGTH 3

GAPDH_counts_length_3 = ( 2, 7, 3, 9 , 0 )
VEGFA_counts_length_3 = ( 0, 6, 2, 2, 0 )
MYC_counts_length_3   = ( 0, 1, 1, 1, 0 )

expected_counts_length_3 = np.array(( GAPDH_counts_length_3, VEGFA_counts_length_3, 
                             MYC_counts_length_3) )

ACTUAL_START_SITE_COVERAGE_length_3 = np.array( [
                                [3, 2, 1, 1, 0], 
                                [0, 3, 2, 1, 0],
                                [0, 0, 0, 0, 0]] )

ACTUAL_STOP_SITE_COVERAGE_length_3 = np.array( [ 
                                [4, 2, 2, 1, 0],
                                [0, 0, 2, 0, 0],
                                [0, 0, 1, 0, 0]] )

#########################################

## LENGTH 4

GAPDH_counts_length_4 = ( 0, 10, 6, 9, 0 )
VEGFA_counts_length_4 = ( 0, 9, 3, 0, 0 )
MYC_counts_length_4   = ( 0, 0, 2, 3, 0 )

expected_counts_length_4 = np.array(( GAPDH_counts_length_4, VEGFA_counts_length_4, 
                             MYC_counts_length_4))

ACTUAL_START_SITE_COVERAGE_length_4 = np.array( [
                                [2, 3, 4, 0, 1], 
                                [0, 4, 3, 2, 0],
                                [0, 0, 0, 0, 0]] )

ACTUAL_STOP_SITE_COVERAGE_length_4 = np.array( [ 
                                        [4, 2, 2, 1, 0],
                                        [0, 0, 0, 0, 0],
                                        [0, 0, 3, 0, 0]] )

#########################################

## LENGTH 5

GAPDH_counts_length_5 = ( 1, 2, 2, 1, 0 )
VEGFA_counts_length_5 = ( 0, 1, 4, 2, 0 )
MYC_counts_length_5   = ( 0, 1, 0, 7, 0 )

expected_counts_length_5 = np.array(( GAPDH_counts_length_5, VEGFA_counts_length_5, 
                             MYC_counts_length_5))

ACTUAL_START_SITE_COVERAGE_length_5 = np.array( [
                                [1, 0, 1, 0, 0], 
                                [0, 1, 0, 0, 0],
                                [0, 0, 0, 0, 0]] )

ACTUAL_STOP_SITE_COVERAGE_length_5 = np.array( [ 
                                        [0, 1, 0, 0, 0],
                                        [0, 0, 2, 0, 0],
                                        [0, 0, 7, 0, 0]] )

#########################################



LEFT_SPAN  = 3
RIGHT_SPAN = 2
METAGENE_RADIUS = 2

ANNOTATION = [\
               ([0, 5], [5, 15], [15, 20]),\
               ([0, 4], [4, 16], [16, 22]),\
               ([0, 3], [3, 10], [10, 17]),\
              ]

#########################################
#####  Expected EXTENDED ANNOTATION 

GAPDH_regions = ( [0, 2], [2, 7], [7, 12], [12, 18], [18, 20] )
VEGFA_regions = ( [0, 1], [1, 7], [7, 13], [13, 19], [19, 22] )
MYC_regions   = ( [0, 0], [0, 6], [6, 7],  [7, 13],  [13,17] )

extended_boundary_regions = ( GAPDH_regions, VEGFA_regions,
                               MYC_regions )
#########################################

METADATA_EXPERIMENT_DICT = {
    "cell_line": "HeLa",
    "digestion_enzyme": "HindIII",
    "digestion_duration": "5 min",
    "external_link" : "https://www.encodeproject.org/",
    "link" : "https://www.encodeproject.org/"
}

METADATA_EXPERIMENT_STR_1 =\
"""cell_line: HeLa
digestion_enzyme: HindIII
digestion_duration: 5 min
link: https://www.encodeproject.org/"""

RIBO_METADATA_STR_1 = \
"""aligner: bowtie2
mapq_threshold : 3
deduplicated: True"""


############################################
############################################

READ_SET_2=\
"""GAPDH 8  11  len_3_CDS_1 0   +
GAPDH 8  11  len_3_CDS_2 0   +
GAPDH 8  11  len_3_CDS_3 0   +
GAPDH 8  11  len_3_CDS_4 0   +
GAPDH 8  11  len_3_CDS_5 0   +
VEGFA 10  13  len_3_CDS_1 0   +
VEGFA 10  13  len_3_CDS_2 0   +
VEGFA 10  13  len_3_CDS_3 0   +
MYC 0  3 len_3_UTR5_junc_1 0   +
MYC 0  3 len_3_UTR5_junc_2 0   +
MYC 0  3 len_3_UTR5_junc_3 0   +
MYC 0  3 len_3_UTR5_junc_4 0   +"""

# Expected Counts for 
## LENGTH 3

READ_SET_2_GAPDH_counts_length_3 = ( 0, 0, 5, 0, 0 )
READ_SET_2_VEGFA_counts_length_3 = ( 0, 0, 3, 0, 0 )
READ_SET_2_MYC_counts_length_3   = ( 4, 0, 0, 0, 0 )

expected_counts_length_3 = np.array(( GAPDH_counts_length_3, VEGFA_counts_length_3, 
                             MYC_counts_length_3) )



def _get_transcripts( file_in_string ):
    rows = file_in_string.split("\n")
    pairs = tuple( map( lambda x: x.split(), rows ) )
    ref_names = tuple( map( lambda x: x[0], pairs ) )
    ref_lengths = tuple( map( lambda x: x[1], pairs ) )
    ref_lengths = tuple (map( int, ref_lengths ) )
    return (ref_names, ref_lengths)

###############################################

RNASEQ_DATA_1 = \
"GAPDH\t2.89\nVEGFA\t15.46\nMYC\t8"

RNASEQ_DATA_2 = \
"GAPDH\t6.42\nVEGFA\t1.37\nMYC\t0.06"


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import re

from setuptools import setup, find_packages


classifiers = [\
    "Development Status :: 2 - Pre-Alpha",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7"
]


def _read(*parts, **kwargs):
    filepath = os.path.join(os.path.dirname(__file__), *parts)
    encoding = kwargs.pop('encoding', 'utf-8')
    with io.open(filepath, encoding=encoding) as fh:
        text = fh.read()
    return text

def get_version():
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        _read('ribogadgets', '_version.py'),
        re.MULTILINE).group(1)
    return version

def get_long_description():
    return _read('README.md')


install_requires = [
    'numpy>=1.16.1',
    'pandas>=0.24.1',
    'click==7.0',
    'h5py>=2.9.0',
    'matplotlib>=3.0.2',
    'multiprocess>=0.70.7',
    'pysam>=0.15.0',
    'pyyaml==5.1'
]

'''
## TODO complete documentation\
extras_require = {
    'docs': [
        'Sphinx>=1.1',
        'numpydoc>=0.5'
    ]
}
'''


setup(
    name='ribogadgets',
    author='Hakan Ozadam',
    version=get_version(),
    packages=find_packages(),
    license='MIT',
    description='Hierarchichal data format based on hdf5 for ribosome profiling data',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=install_requires,
    zip_safe=False,
    classifiers=[x.strip() for x in classifiers ],
    entry_points='''
        [console_scripts]
        ribog=ribogadgets.cli.main:cli
    '''
)



# -*- coding: utf-8 -*-

def read_line_chunks( file_object, max_bytes ):
    """
    Function that lazily reads file in chunks of lines
    it reads upto max_bytes and then till the end of the existing line
    returns the aray of lines read 
    """
    while True:
        data = file_object.readlines( max_bytes)
        if not data:
            break
        yield data


# -*- coding: utf-8 -*-
import unittest
from unittest import mock
from unittest.mock import patch
import os
from io import StringIO
import subprocess

import numpy as np
import h5py

from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *

from multilength_test_data import *

###########################################

NPROCESS = 4

ribo_output_file = "cli_create.ribo"


class TestCLIBase(unittest.TestCase):

    def run_command(self, command):
      """
      Command must be a list.
      See examples below,
      """

      command_str    = tuple( map( str, command ) )
      process        = subprocess.Popen(command_str, 
                                        stdout = subprocess.PIPE,
                                        stderr = subprocess.PIPE)
      output, error  = process.communicate()
      output_str     = output.decode()
      error_str      = error.decode()

      return (output_str, error_str) 

    def setUp(self):
        self.tmp_files = list()

        files_and_contents = \
            [ (REF_LEN_FILE,     TRANSCRIPT_LENGTHS),
              (ANNOTATION_FILE,  TRANSCRIPT_ANNOTATION),
              (ALIGNMENT_FILE_1, READ_SET_1),
              (METADATA_FILE_1,  METADATA_EXPERIMENT_STR_1),
              (RIBO_META_FILE_1, RIBO_METADATA_STR_1),
              (RNASEQ_FILE_1,    RNASEQ_DATA_1) ]

        for t_file, content_str in files_and_contents:
            with open(t_file, "w") as output_stream:
                print(content_str, file = output_stream)
            self.tmp_files.append(t_file)

        if os.path.isfile(ribo_output_file):
            os.remove(ribo_output_file)

        create_command = ["ribog", "create", 
                          "--alignmentfile",  ALIGNMENT_FILE_1,
                          "--name",          "merzifon", 
                          "--reference",     "hg38",
                          "--lengths",        REF_LEN_FILE, 
                          "--annotation",     ANNOTATION_FILE,
                          "--metageneradius", METAGENE_RADIUS, 
                          "--leftspan",       LEFT_SPAN, 
                          "--rightspan",      RIGHT_SPAN,
                          "--lengthmin",      2, 
                          "--lengthmax",      5, 
                          "--expmeta",        METADATA_FILE_1,
                          "--ribometa",       RIBO_META_FILE_1,
                          "--rnaseq",         RNASEQ_FILE_1,
                          "-n",               4, 
                          ribo_output_file]

        output, error      = self.run_command(create_command)

        #print("Output:", output)
        self.tmp_files.append( ribo_output_file )
        if output:
            print(output)
        if error:
            print(error)


    def tearDown(self):
        [ os.remove(f) for f in self.tmp_files ]

#######################################################################        


OUT_PDF = "out.pdf"

class TestCLIPlot(TestCLIBase):

    """
    def tearDown(self):
        TestCLIBase.tearDown(self)
        if os.path.exists(OUT_PDF):
            os.remove(OUT_PDF)
    """


    def test_plot_metagene_start(self):
        command_pieces = ["ribog", "plot", "metagene",
                          "--lowerlength", "2", 
                          "--upperlength", "3", 
                          "-o",            OUT_PDF,
                          "--site",        "start", 
                          ribo_output_file, "merzifon"]
        output, error = self.run_command(command_pieces)
        self.assertTrue(os.path.exists(OUT_PDF) )


    def test_plot_lengthdist(self):
        command_pieces = ["ribog", "plot", "lengthdist", 
                          "-o",        OUT_PDF, 
                          "--title",  "title_here",
                          "--region", "CDS", 
                          ribo_output_file, 
                          "merzifon"]

        output, error = self.run_command(command_pieces)

        self.assertTrue(os.path.exists(OUT_PDF) )

    def test_plot_region_counts(self):
        command_pieces = ["ribog", "plot", "regioncounts",
                          "-o",       OUT_PDF, 
                          "--title",  "title_here",
                          ribo_output_file, 
                          "merzifon"]
        output, error = self.run_command(command_pieces)
        self.assertTrue(os.path.exists(OUT_PDF) )


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-

import h5py

from .create import create_ribo
from .settings import *

########################################################
###   A P I   F O R   R I B O G A D G E T S
########################################################

# INCOMPLETE

class ribo:
    def __init__(self, file_path):
        self.handle = h5py.File(file_path)

    @property
    def experiments(self):
        return self.handle[EXPERIMENTS_name].keys()


    def __del__(self):
        self.handle.close()


# -*- coding: utf-8 -*-

from collections import OrderedDict
import datetime

import yaml

from ..settings import *
from ..core.get_gadgets import get_experiment_names
from .file import flex_open

##################################################3


def metadata_dict_to_aligned_str(metadata_dict):
    """
    Aligns metadata keys.

    Returns the resulting string.
    """

    if not metadata_dict or len(tuple( metadata_dict.keys() )) == 0:
        return ""

    max_key_len = max( map(len, metadata_dict.keys()) )
    
    def aligner(x):
        return  x + ( " " * ( max_key_len - len(x) ) )\
                  + " : " + str(metadata_dict[x])
    
    aligned_keys = tuple(map( aligner, metadata_dict.keys()))
    
    return "\n".join(aligned_keys)



### Belof this line is to be revised or depracated
def read_metadata_from_yaml(metadata):
    with flex_open(metadata) as metadata_input:
        result_dict = yaml.load(metadata_input, Loader=yaml.FullLoader)
    return result_dict

"""
# This function is deprecated
# Remove this 
def sanitize_metadata(metadata_dict):
    clean_metadata = OrderedDict()
    essential_keys = LIBRARY_ESSENTIAL_METADATA.keys()
    invalid_keys = []

    for k in essential_keys:
        clean_metadata[k] = ""

    for s in metadata_dict.keys():
        k = s.lower()
        if k not in essential_keys:
            invalid_keys.append(k)
        else:
            clean_metadata[k] = str(metadata_dict[s])

    return clean_metadata, invalid_keys
"""    

def get_ribo_file_metadata_for_display(ribo_handle):
    ribo_metadata = OrderedDict()

    for key, value in RIBO_METADATA_FOR_DISPLAY.items():
        ribo_metadata[value] = ribo_handle.attrs[key]

    # format time
    time_in_seconds = ribo_metadata[RIBO_METADATA_FOR_DISPLAY[ATTRS_TIME]] 
    val = datetime.datetime.fromtimestamp(time_in_seconds)
    ribo_metadata[RIBO_METADATA_FOR_DISPLAY[ATTRS_TIME]] = \
        val.strftime('%Y-%m-%d %H:%M:%S') 

    return ribo_metadata









# -*- coding: utf-8 -*-
import unittest
import os
from io import StringIO
import subprocess

import numpy as np
import h5py

from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *
from ribogadgets.core.quantify import quantify_experiment

from multilength_test_data import *
from ribogadgets.core import create_experiment

###########################################

NPROCESS = 4

ribo_output_file = "cli_rnaseq.ribo"


class TestCLIBase(unittest.TestCase):

    def run_command(self, command):
      """
      Command must be a list.
      See examples below,
      """

      command_str    = tuple( map( str, command ) )
      process        = subprocess.Popen(command_str, 
                                        stdout = subprocess.PIPE,
                                        stderr = subprocess.PIPE)
      output, error  = process.communicate()
      output_str     = output.decode()
      error_str      = error.decode()

      return (output_str, error_str) 



    def setUp(self):
        self.tmp_files = list()

        files_and_contents = \
            [ (REF_LEN_FILE,     TRANSCRIPT_LENGTHS),
              (ANNOTATION_FILE,  TRANSCRIPT_ANNOTATION),
              (ALIGNMENT_FILE_1, READ_SET_1),
              (RNASEQ_FILE_1,    RNASEQ_DATA_1),
              (RNASEQ_FILE_2,    RNASEQ_DATA_2) ]

        for t_file, content_str in files_and_contents:
            with open(t_file, "w") as output_stream:
                print(content_str, file = output_stream)
            self.tmp_files.append(t_file)

        if os.path.isfile(ribo_output_file):
            os.remove(ribo_output_file)

        create_command = ["ribog", "create",
                          "--alignmentfile",  ALIGNMENT_FILE_1,  
                          "--name",           "merzifon",
                          "--reference",      "hg38",
                          "--lengths",        REF_LEN_FILE, 
                          "--annotation",     ANNOTATION_FILE,
                          "--metageneradius", METAGENE_RADIUS, 
                          "--leftspan",       LEFT_SPAN, 
                          "--rightspan",      RIGHT_SPAN,
                          "--lengthmin",      2, 
                          "--lengthmax",      5, 
                          "-n",               4,
                          ribo_output_file]

        output, error = self.run_command(create_command)
        if error:
            print("Error in creation of ribo:\n", error)
        self.tmp_files.append( ribo_output_file )


    def tearDown(self):
        [ os.remove(f) for f in self.tmp_files ]

#######################################################################  


class TestCLIRNASEQ(TestCLIBase):

    #@unittest.skip("temporarily skipping plot tests")
    def test_dump_annotation(self):
        command_pieces  = ["ribog", "dump", "annotation", ribo_output_file]
        output, error   = self.run_command(command_pieces)
        output_lines    = output.split("\n")
        expected_line_2 = ["GAPDH", "5", "15", "CDS", "0", "+"]
        output_line2    = output_lines[1].split()
        self.assertTrue( expected_line_2 == output_line2 )

    def test_set_rnaseq(self):
        command_pieces = ["ribog", "rnaseq", "set",
                          "--name",   "merzifon",
                          "--rnaseq", RNASEQ_FILE_1,
                           ribo_output_file]

        output, error   = self.run_command(command_pieces)
        # Check that we don't get error messages
        self.assertEqual(output, "")
        self.assertEqual(error, "")


        # Check that --force option oworks when there is
        # existing rna-seq

        command_pieces = ["ribog", "rnaseq", "set",
                          "--name",   "merzifon",
                          "--rnaseq", RNASEQ_FILE_1,
                          "--force",
                           ribo_output_file]

        output, error   = self.run_command(command_pieces)

    #@unittest.skip("temporarily skipping plot tests")
    def test_get_rnaseq(self):
        command_pieces = ["ribog", "rnaseq", "set",
                          "--name",   "merzifon",
                          "--rnaseq", RNASEQ_FILE_1,
                           ribo_output_file]

        output, error   = self.run_command(command_pieces)

        command_pieces = ["ribog", "rnaseq", "get",
                          "--name", "merzifon",
                           ribo_output_file]

        output, error   = self.run_command(command_pieces)

        output_lines          = output.split("\n")
        VEGFA_name, VEGFA_exp = output_lines[2].split()

        self.assertEqual(VEGFA_name, "VEGFA")
        self.assertTrue(np.isclose( float(VEGFA_exp), 15.46 ))


    #@unittest.skip("temporarily skipping plot tests")
    def test_delete_rnaseq(self):
        command_pieces = ["ribog",    "rnaseq", "set",
                          "--name",   "merzifon",
                          "--rnaseq", RNASEQ_FILE_1,
                           ribo_output_file]

        output, error   = self.run_command(command_pieces)

        command_pieces = ["ribog",    "rnaseq", "delete",
                          "--name",   "merzifon",
                          "--force",
                           ribo_output_file]

        output, error   = self.run_command(command_pieces)
        print(output, error)

        ribo = h5py.File(ribo_output_file)

        self.assertTrue( RNASEQ_name not in \
                         ribo[EXPERIMENTS_name]["merzifon"].keys() )

        ribo.close()


if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-
import unittest
from unittest import mock
from unittest.mock import patch
import os
from io import StringIO, BytesIO

import numpy as np
import h5py

from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *
from ribogadgets.core.aggregate_by_length import *

from multilength_test_data import *
from ribogadgets.core import create_experiment

###########################################

NPROCESS = 4

class TestCreate(unittest.TestCase):
    def setUp(self):

        self.len_file        = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file = StringIO(TRANSCRIPT_ANNOTATION)
        self.alignment_file  = StringIO(READ_SET_1)

        self.handle = h5py.File(BytesIO())

        create.create_ribo(self.handle, 
                experiment_name        = "merzifon", 
                alignment_file      = self.alignment_file,
                reference_name      = "hg38",
                lengths_file        = self.len_file, 
                annotation_file     = self.annotation_file,
                metagene_radius     = METAGENE_RADIUS, 
                left_span = LEFT_SPAN, right_span = RIGHT_SPAN,
                length_min = 2, length_max = 5, 
                tmp_file_prefix     = "")
        
    def tearDown(self):
        self.handle.close()


    ### T E S T    S U M S  ################################    

    def test_aggregate_region_sum(self):
        aggregate_2_3_region = aggregate_region_counts(
                                 self.handle, range_lower = 2, range_upper =3,
                                 sum_values = True )
        expected_reg_count_len_2_3 = expected_counts_length_2 + \
                                       expected_counts_length_3

        # print(aggregate_2_3_region, "\n-----\n" , expected_reg_count_len_2_3 )
        self.assertTrue( np.all(aggregate_2_3_region ==\
                                expected_reg_count_len_2_3) )

        expected_reg_count_len_2_3_4_5 = expected_counts_length_2 + \
                                       expected_counts_length_3 + \
                                       expected_counts_length_4 + \
                                       expected_counts_length_5

        aggregate_2_5_region = aggregate_region_counts(
                                 self.handle, range_lower = 2, range_upper =5,
                                 sum_values = True )

        self.assertTrue( np.all(expected_reg_count_len_2_3_4_5 ==\
                                aggregate_2_5_region) )

        aggregate_2_region = aggregate_region_counts(
                                 self.handle, range_lower = 2, range_upper =2,
                                 sum_values      = True,
                                 experiment_list = ["merzifon", "merzifon"] )
        self.assertTrue( np.all(aggregate_2_region ==\
                                expected_counts_length_2) )


    def test_aggregate_start_site_sum(self):
        aggregate_2_3_start = aggregate_start_site(
                                 self.handle, range_lower = 2, range_upper =3,
                                 sum_values = True )
        expected_2_3_start = ACTUAL_START_SITE_COVERAGE_length_2 + \
                             ACTUAL_START_SITE_COVERAGE_length_3

        self.assertTrue( np.all(aggregate_2_3_start == expected_2_3_start)  )

        aggregate_2_start = aggregate_start_site(
                                 self.handle, range_lower = 2, range_upper =2,
                                 sum_values = True )

        self.assertTrue( np.all(aggregate_2_start ==\
                                ACTUAL_START_SITE_COVERAGE_length_2)  )

        aggregate_2_to_5_start = aggregate_start_site(
                                 self.handle, range_lower = 2, range_upper =5,
                                 sum_values = True )

        expected_2_to_5_start = ACTUAL_START_SITE_COVERAGE_length_2 + \
                             ACTUAL_START_SITE_COVERAGE_length_3 + \
                             ACTUAL_START_SITE_COVERAGE_length_4 + \
                             ACTUAL_START_SITE_COVERAGE_length_5

        self.assertTrue( np.all(aggregate_2_to_5_start == \
                                expected_2_to_5_start)  )


    def test_aggregate_stop_site_sum(self):
        aggregate_2_3_stop = aggregate_stop_site(
                                 self.handle, range_lower = 2, range_upper =3,
                                 sum_values = True )

        expected_2_3_stop = ACTUAL_STOP_SITE_COVERAGE_length_2 + \
                             ACTUAL_STOP_SITE_COVERAGE_length_3

        self.assertTrue( np.all(aggregate_2_3_stop == expected_2_3_stop)  )


    ### Test GROUPINGS #######################
    def test_aggregate_region_group(self):
        aggregate_2_3_region = aggregate_region_counts(
                                 self.handle, range_lower = 2, range_upper =3,
                                 sum_values = False )

        length_col  = np.repeat( range(2,3+1), len( expected_counts_length_2 ) )
        length_col =  np.array( length_col, ndmin=2 )

        combined_counts = np.concatenate( (expected_counts_length_2,
                                           expected_counts_length_3) )

        expected_array = np.concatenate( (length_col.T, combined_counts), axis = 1 )

        array_comparison = (expected_array == aggregate_2_3_region)
        self.assertTrue( np.all(array_comparison) )

    def test_aggregate_start_site_group(self):
        aggregate_4_5_start = aggregate_start_site(
                                 self.handle, range_lower = 4, range_upper =5,
                                 sum_values = False )

        length_col  = np.repeat( range(4,5+1), 
                                 len( ACTUAL_START_SITE_COVERAGE_length_4 ) )
        length_col =  np.array( length_col, ndmin=2 )

        combined_counts = np.concatenate( (ACTUAL_START_SITE_COVERAGE_length_4,
                                           ACTUAL_START_SITE_COVERAGE_length_5) )

        expected_array = np.concatenate( (length_col.T, combined_counts), axis = 1 )
        array_comparison = (expected_array == aggregate_4_5_start)
        self.assertTrue( np.all(array_comparison) )

    def test_aggregate_stop_site_group(self):
        aggregate_2_5_stop = aggregate_stop_site(
                                 self.handle, range_lower = 2, range_upper =5,
                                 sum_values = False )

        length_col  = np.repeat( range(2,5+1), len( ACTUAL_STOP_SITE_COVERAGE_length_4 ) )
        length_col =  np.array( length_col, ndmin=2 )

        combined_counts = np.concatenate( (ACTUAL_STOP_SITE_COVERAGE_length_2,
                                           ACTUAL_STOP_SITE_COVERAGE_length_3,
                                           ACTUAL_STOP_SITE_COVERAGE_length_4,
                                           ACTUAL_STOP_SITE_COVERAGE_length_5) )

        expected_array = np.concatenate( (length_col.T, combined_counts), axis = 1 )
        array_comparison = (expected_array == aggregate_2_5_stop)
        self.assertTrue( np.all(array_comparison) )



if __name__ == '__main__':
        
    unittest.main()